import asyncio
import traceback
from datetime import datetime
from typing import Callable
from .timers import Timer
from .event import ScheduleTaskEvent
from .schedule import EventSystem
from ...letoderea import Subscriber
from ...letoderea.handler import await_exec_target

class TimingTask:
    callable_target: Callable
    task: asyncio.Task
    timer: Timer
    event_system: EventSystem
    is_disposable: bool = False
    is_stop: bool = False
    is_start: bool = False
    interval_offset: float = 0

    def __init__(
            self,
            callable_target: Callable,
            timer: Timer,
            event_system: EventSystem,
            is_disposable: bool = False,
    ) -> None:
        self.callable_target = callable_target
        self.timer = timer
        self.event_system = event_system
        self.is_disposable = is_disposable

    def set_task(self):
        if not self.is_start:
            self.task = self.event_system.loop.create_task(self.run_task())
        else:
            raise RuntimeError("the scheduler task has been started!")

    def interval_generator(self):
        for next_execute_time in self.timer.get_delta():
            if self.is_stop:
                return
            now = datetime.now()
            if next_execute_time >= now:
                yield (next_execute_time - now).total_seconds() - self.interval_offset

    def coroutine_generator(self):
        for sleep_interval in self.interval_generator():
            yield asyncio.sleep(sleep_interval), True
            yield await_exec_target(
                Subscriber.set()(self.callable_target), ScheduleTaskEvent.get_params
            ), False

    async def run_task(self) -> None:
        for coro, waiting in self.coroutine_generator():
            if waiting:  # 是否为 asyncio.sleep 的 coro
                try:
                    await coro
                except asyncio.CancelledError:
                    return
            else:  # 执行
                try:
                    await coro
                except Exception as e:
                    traceback.print_exc()
                else:
                    now = datetime.now()
                    self.interval_offset = (now.microsecond / 1000000)
                    if self.is_disposable:
                        self.is_stop = True

    async def join(self, stop=False):
        if stop and not self.is_stop:
            self.is_stop = True

        if self.task:
            await self.task

    def stop(self):
        if not self.task.cancelled():
            self.task.cancel()



import functools
from contextvars import ContextVar
from contextlib import contextmanager
from enum import Enum
from typing import Callable, Any, Union, TYPE_CHECKING, Dict, Type

from pydantic import BaseModel, BaseConfig, Extra

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny

bot_application = ContextVar("bot_application")
event = ContextVar("event")
event_loop = ContextVar("event_loop")
event_system = ContextVar("event_system")
upload_method = ContextVar("upload_method")

from .exceptions import (
    AccountMuted,
    AccountNotFound,
    InvalidArgument,
    InvalidSession,
    InvalidVerifyKey,
    MessageTooLong,
    UnknownError,
    UnknownTarget,
    UnVerifiedSession,
)

code_exceptions_mapping: Dict[int, Type[Exception]] = {
    1: InvalidVerifyKey,
    2: AccountNotFound,
    3: InvalidSession,
    4: UnVerifiedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: MessageTooLong,
    400: InvalidArgument
}


def error_check(code: Union[dict, int]):
    origin = code
    if isinstance(code, dict):
        code = code.get("code")
    else:
        code = code
    if not isinstance(code, int) or code == 200 or code == 0:
        return
    exc_cls = code_exceptions_mapping.get(code)
    if exc_cls:
        raise exc_cls(exc_cls.__doc__)
    else:
        raise UnknownError(f"{origin}")


class ContextModel:
    content: Any

    @classmethod
    def set(cls, ct):
        cls.content = ct

    @classmethod
    def get(cls):
        return cls.content

    @classmethod
    def reset(cls, ct):
        cls.content = None
        cls.content = ct


class ContextCollection:
    bot_application: ContextModel
    event: ContextModel
    event_loop: ContextModel
    event_system: ContextModel
    upload_method: ContextModel


class Structured(BaseModel):
    """
    一切数据模型的基类.
    """

    def dict(
            self,
            *,
            include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
            exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
            by_alias: bool = False,
            skip_defaults: bool = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
    ) -> "DictStrAny":
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=True,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=True,
        )

    class Config(BaseConfig):
        extra = Extra.allow


class UploadMethods(Enum):
    """用于向 `uploadImage` 或 `uploadVoice` 方法描述图片的上传类型"""

    Friend = "friend"
    Group = "group"
    Temp = "temp"


@contextmanager
def enter_message_send_context(method: UploadMethods):
    t = upload_method.set(method)
    yield
    upload_method.reset(t)


@contextmanager
def enter_context(bot=None, event_i=None):
    t1 = None
    t2 = None
    t3 = None
    t4 = None

    if bot:
        t1 = bot_application.set(bot)
        t3 = event_loop.set(bot.event_system.loop)
        t4 = event_system.set(bot.event_system)
    if event_i:
        t2 = event.set(event_i)

    yield
    try:
        if t1:
            bot_application.reset(t1)

        if all([t2, t3, t4]):
            event.reset(t2)
            event_loop.reset(t3)
            event_system.reset(t4)
    except ValueError:
        pass


def bot_application_context_manager(func: Callable):
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        with enter_context(bot=self):
            return await func(self, *args, **kwargs)

    return wrapper



from typing import List, Iterable, Type, Union

from arclet.cesloi.message.element import MessageElement, _update_forward_refs, Source, Quote, File
from ..utils import Structured


class MessageChain(Structured):
    """
    即 "消息链", 用于承载整个消息内容的数据结构, 包含有一有序列表, 包含有继承了 MessageElement 的类实例.
    """

    __root__: List[MessageElement]

    @staticmethod
    def element_class_generator(target=MessageElement):
        for i in target.__subclasses__():
            yield i
            if i.__subclasses__():
                yield from MessageChain.element_class_generator(i)

    @staticmethod
    def search_element(name: str):
        for i in MessageChain.element_class_generator():
            if i.__name__ == name:
                return i

    @staticmethod
    def build_chain(obj: List[Union[dict, MessageElement]]):
        elements = []
        for i in obj:
            if isinstance(i, MessageElement):
                elements.append(i)
            elif isinstance(i, dict) and "type" in i:
                elements.append(MessageChain.search_element(i["type"]).parse_obj(i))
        return elements

    @classmethod
    def parse_obj(cls: Type["MessageChain"], obj: List[Union[dict, MessageElement]]) -> "MessageChain":
        return cls(__root__=cls.build_chain(obj))  # 默认是不可变型

    def __init__(self, __root__: Iterable[MessageElement]):
        super().__init__(__root__=self.build_chain(list(__root__)))

    @classmethod
    def create(cls, *elements: Union[Iterable[MessageElement], MessageElement, str]) -> "MessageChain":
        element_list = []
        for ele in elements:
            if isinstance(ele, MessageElement):
                element_list.append(ele)
            elif isinstance(ele, str):
                from .element import Plain
                element_list.append(Plain(ele))
            else:
                element_list.extend(list(ele))
        return cls(__root__=element_list)

    def to_text(self) -> str:
        """获取以字符串形式表示的消息链, 且趋于通常你见到的样子.

        Returns:
            str: 以字符串形式表示的消息链
        """
        return "".join(i.to_text() for i in self.__root__)

    def to_serialization(self) -> str:
        """获取可序列化的字符串形式表示的消息链, 会存储所有的信息.

        Returns:
            str: 序列化的字符串形式的消息链
        """
        return "__root__: " + "".join(i.to_serialization()
                                      for i in self.replace_text('[', '[_').replace_text(']', '_]').__root__)

    @staticmethod
    def from_text(text: str) -> "MessageChain":
        from .element import Plain
        return MessageChain([Plain(text)])

    def only_text(self) -> str:
        """获取消息链中的纯文字部分
        """
        return "".join(i.to_text() if i else "" for i in self.findall("Plain"))

    def is_instance(self, element_type: Union[str, Type[MessageElement]]) -> bool:
        if isinstance(element_type, str):
            element_type = MessageChain.search_element(element_type)
        for i, v in enumerate(self.__root__):
            if type(v) in (Source, Quote):
                continue
            if type(v) is not element_type:
                return False
        return True

    def findall(self, element_type: Union[str, Type[MessageElement]]) -> List[MessageElement]:
        """返回消息链内可能的所有指定元素"""
        if isinstance(element_type, str):
            element_type = MessageChain.search_element(element_type)
        return [i for i in self.__root__ if type(i) is element_type]

    def find(self, element_type: Union[str, Type[MessageElement]], index: int = 0) -> Union[bool, MessageElement]:
        """
        当消息链内有指定元素时返回该元素
        无则返回False

        Args:
            element_type : 指定的元素类型
            index: 位置索引, 默认为0
        """
        ele = self.findall(element_type)
        return False if not ele else ele[index]

    def has(self, element_type: Union[str, Type[MessageElement]]) -> bool:
        """
        当消息链内有指定元素时返回True
        无则返回False
        """
        ele = self.findall(element_type)
        return False if not ele else True

    def pop(self, index: int) -> MessageElement:
        return self.__root__.pop(index)

    def index(self, element_type: Union[str, Type[MessageElement]]) -> int:
        ele = self.find(element_type)
        if ele:
            return self.__root__.index(ele)
        else:
            raise ValueError(f"{element_type} is not in this MessageChain")

    def replace(self, element_type: Union[str, Type[MessageElement]],
                new_element: MessageElement, counts: int = None) -> "MessageChain":
        """替换消息链中的所有指定的消息元素类型为新的消息元素实例；不改变消息链本身

        Args:
            element_type : 要替换的元素的类型
            new_element: 新的消息元素实例
            counts: 替换的次数,不填写时默认为替换全部符合的元素
        Returns:
            MessageChain: 新的消息链
        """
        if isinstance(element_type, str):
            element_type = MessageChain.search_element(element_type)
        new_message = MessageChain(self.__root__)
        if not counts:
            for i, v in enumerate(self.__root__):
                if type(v) is element_type:
                    new_message.__root__[i] = new_element
        elif counts > 0:
            for i, v in enumerate(self.__root__):
                if type(v) is element_type:
                    counts -= 1
                    new_message.__root__[i] = new_element
                if counts == 0:
                    break
        return new_message

    def replace_text(self, old_text: str, new_text: str, counts: int = -1) -> "MessageChain":
        """替换消息链中可能含有的文本消息中的文本为指定文本；不改变消息链本身

        Args:
            old_text : 要替换的文本内容
            new_text: 新的文本内容
            counts: 替换的次数,不填写时默认为替换全部符合的元素
        Returns:
            MessageChain: 新的消息链
        """
        new_message = MessageChain(self.__root__)
        from arclet.cesloi.message.element import Plain
        for ele in new_message:
            if isinstance(ele, Plain):
                ele.text = ele.text.replace(old_text, new_text, counts)
        return new_message

    def remove(self, element_type: Union[str, Type[MessageElement]], counts: int = None):
        """删除消息链中的所有指定的消息元素类型

        Args:
            element_type : 要删除的元素的类型
            counts: 删除的次数,不填写时默认为删除全部符合的元素
        Returns:
            操作完成后的消息链本身
        """
        if isinstance(element_type, str):
            element_type = MessageChain.search_element(element_type)
        if not counts:
            self.__root__ = [i for i in self.__root__ if type(i) is not element_type]
        elif counts > 0:
            i = 0
            while counts:
                if type(self.__root__[i]) is element_type:
                    try:
                        self.__root__.remove(self.__root__[i])
                        counts -= 1
                        i -= 1
                    except ValueError:
                        break
                i += 1
        return self

    def only_save(self, element_type: Union[str, Type[MessageElement]]):
        """删除消息链中的所有非指定的消息元素类型

        Args:
            element_type : 要保留的元素的类型
        Returns:
            操作完成后的消息链本身
        """
        if isinstance(element_type, str):
            element_type = MessageChain.search_element(element_type)
        self.__root__ = [i for i in self.__root__ if type(i) is element_type]
        return self

    def insert(self, index: int, element: MessageElement):
        """在指定位置插入一个消息元素实例
        Args:
            index: 插入的位置
            element: 需要插入的消息元素
        Returns:
            操作完成后的消息链本身
        """
        self.__root__.insert(index, element)
        return self

    def append(self, element: MessageElement):
        """在消息链尾部增加一个消息元素实例

        Args:
            element: 需要插入的消息元素
        Returns:
            操作完成后的消息链本身
        """
        self.__root__.append(element)
        return self

    def extend(self, *elements: Union[MessageElement, List[MessageElement]]):
        element_list = []
        for ele in elements:
            if isinstance(ele, MessageElement):
                element_list.append(ele)
            else:
                element_list.extend(ele)
        self.__root__ += element_list
        return self

    def copy_self(self) -> "MessageChain":
        return MessageChain(self.__root__)

    def __add__(self, other) -> "MessageChain":
        if isinstance(other, MessageElement):
            self.__root__.append(other)
            return self
        elif isinstance(other, MessageChain):
            self.__root__.extend(i for i in other.__root__ if i.type != "Source")
            return self
        elif isinstance(other, List):
            self.__root__ += other
            return self

    def __repr__(self) -> str:
        return fr"MessageChain({repr(self.__root__)})"

    def __iter__(self) -> Iterable[MessageElement]:
        yield from self.__root__

    def __getitem__(self, index) -> MessageElement:
        return self.__root__[index]

    def __len__(self) -> int:
        return len(self.__root__)

    def __contains__(self, item: Union[Type[MessageElement], str]) -> bool:
        """
        是否包含特定元素类型/字符串
        """
        if isinstance(item, str):
            return item in self.find("Plain").to_text()
        else:
            return self.has(item)

    def to_sendable(self):
        return self.remove(Source).remove(Quote).remove(File)

    def startswith(self, string: str) -> bool:
        if not self.__root__:
            return False
        return self.to_text().startswith(string)

    def endswith(self, string: str) -> bool:
        if not self.__root__:
            return False
        return self.to_text().endswith(string)


_update_forward_refs()

if __name__ == "__main__":
    from arclet.cesloi.message.element import Plain, At, Image
    msg = MessageChain.create(At(123, display="ccc"), Plain("[bbb]"), Image(base64="aaaa"))
    print(str(msg))
    print(msg.to_text())
    print(msg.to_serialization())
    print(msg.append(At(456)).insert(2, Plain(" ddd")).to_text())



import asyncio
from datetime import datetime, timedelta
from typing import Optional, List
from .timers import Timer

from ...letoderea import EventSystem
from ...letoderea.entities.condition import TemplateCondition
from .task import TimingTask


class Toconada:
    """
    托可娜达,莱恩家的姐姐，与妹妹一起住在Cesloi隔壁

    用于非事件订阅器的调度器

    Args:
        event_system: 事件系统的实例
    """
    event_system: EventSystem
    loop: asyncio.AbstractEventLoop
    schedule_tasks: List[TimingTask]

    def __init__(self, event_system: EventSystem):
        self.event_system = event_system
        self.loop = self.event_system.loop
        self.schedule_tasks = []

    def timing(self, timer: Timer, is_disposable: Optional[bool] = False):
        """定时一个函数/方法

        Args:
            timer : 时间器实例, 参考timing.timers
            is_disposable: 是否只执行一次该函数/方法
        """
        def wrapper(func):
            task = TimingTask(
                func, timer, self.event_system, is_disposable
            )
            self.schedule_tasks.append(task)
            task.set_task()
            return func

        return wrapper


class Toconado(TemplateCondition):
    """
    托可娜多,莱恩家的妹妹,与Cesloi玩得更好

    用于事件订阅器的调度器

    Args:
            timer : 时间器实例, 参考timing.timers
    """
    timer: Timer
    last_run: Optional[datetime]

    def __init__(self, timer: Timer):
        self.timer = timer
        self.last_run = None

    def judge(self) -> bool:
        if not self.last_run:
            self.last_run = datetime.now()
            return True
        for interval in self.timer.get_delta():
            now = datetime.now()
            if self.last_run + (interval - now) <= now:
                self.last_run = now.replace(microsecond=0)
                return True
            return False



import aiohttp
from pathlib import Path
from .relation import Group, Friend
from typing import Optional, Literal, Union
from pydantic import BaseModel


class BotMessage(BaseModel):
    messageId: int


class Profile(BaseModel):
    nickname: str
    email: Optional[str]
    age: Optional[int]
    level: int
    sign: str
    sex: Literal["UNKNOWN", "MALE", "FEMALE"]


class DownloadInfo(BaseModel):
    """群组文件的下载信息"""
    download_times: Optional[int]
    uploader_id: Optional[int]
    upload_time: Optional[int]
    last_modify_time: Optional[int]

    url: Optional[str] = None
    sha1: Optional[str] = None
    md5: Optional[str] = None


class FileInfo(BaseModel):
    """描述群组文件的有关状态"""

    name: Optional[str] = None
    id: Optional[str] = None
    path: Optional[str] = None
    parent: Optional["FileInfo"] = None
    contact: Optional[Union[Group, Friend]] = None
    is_file: Optional[bool] = None
    is_directory: Optional[bool] = None
    downloadInfo: Optional[DownloadInfo] = None

    async def save_file(self, save_path: Union[Path, str]) -> None:
        """
        该方法不返回文件的二进制数据
        """
        if self.is_file:
            if isinstance(save_path, str):
                save_path = Path(save_path)
            if not self.downloadInfo:
                raise AttributeError("cannot download")
            try:
                with save_path.open("wb") as f_obj:
                    async with aiohttp.request("GET", self.downloadInfo.url) as resp:
                        async for result in resp.content:
                            f_obj.write(result)
            except Exception as e:
                save_path.unlink()
                raise e
        else:
            return


FileInfo.update_forward_refs(FileInfo=FileInfo)



import asyncio
import inspect
import json
import random
from asyncio import Task

import aiohttp
from typing import Optional, Union, Dict, TYPE_CHECKING, Awaitable, Callable
from aiohttp import ClientSession, WSMsgType
from yarl import URL

from arclet.cesloi.utils import enter_context, Structured
from arclet.letoderea import EventSystem, search_event
from arclet.cesloi.logger import Logger
from .utils import error_check
from .event.base import MiraiEvent

if TYPE_CHECKING:
    from .bot_client import Cesloi


class BotSession(Structured):
    host: str
    single_mode: bool = False
    account: Optional[int] = None
    verifyKey: Optional[str] = None
    sessionKey: Optional[str] = None
    version: Optional[str] = None

    def __init__(
            self,
            host: str,
            account: Optional[int] = None,
            verify_key: Optional[str] = None,
            *,
            single_mode: bool = False,
    ) -> None:
        """
            用于描述与上游接口会话, 并存储会话状态的实体类.

            Args:
                host (AnyHttpUrl): `mirai-api-http` 服务所在的根接口地址
                account (int): 应用所使用账号的整数 ID, singleMode 模式下为空, 非 singleMode 下新建连接不可为空.
                verify_key (str): 配置文件中指定,用以与mah绑定账号
            """
        super().__init__(
            host=host, account=account, verifyKey=verify_key, single_mode=single_mode
        )


class Communicator:
    bot_session: BotSession
    event_system: EventSystem
    loop: asyncio.AbstractEventLoop
    logger: Logger.logger

    def __init__(
            self,
            bot_session: BotSession,
            bot: "Cesloi",
            event_system: EventSystem,
            logger: Optional[Logger] = None
    ):
        self.bot_session = bot_session
        self.event_system = event_system
        self.loop = event_system.loop
        self.logger = logger or Logger().logger
        self.bot = bot
        self.running_task: Optional[Task] = None
        self.running: bool = False
        self.ws_connection: Optional[aiohttp.ClientWebSocketResponse] = None
        self.client_session: Optional[ClientSession] = None
        self.wait_response_future: Dict[str, asyncio.Future] = {}
        self.timeout: float = 60.0

    async def stop(self):
        self.running = False
        if self.running_task and not self.running_task.done():
            try:
                await self.running_task
            except asyncio.CancelledError:
                pass
        self.running_task = None
        self.bot_session.sessionKey = None
        await self.client_session.close()

    @staticmethod
    async def run_always_await(any_callable: Union[Awaitable, Callable]):
        if inspect.isawaitable(any_callable):
            return await any_callable
        else:
            return any_callable

    async def parse_to_event(self, data: dict) -> MiraiEvent:
        """
        从尚未明确指定事件类型的对象中获取事件的定义, 并进行解析

        Args:
            data (dict): 用 dict 表示的序列化态事件, 应包含有字段 `type` 以供分析事件定义.

        Returns:
            MiraiEvent: 已经被序列化的事件
        """
        event_type: Optional[str] = data.get("type")
        if not event_type or not isinstance(event_type, str):
            raise TypeError("Unable to find 'type' field for automatic parsing")
        event_class: Optional[MiraiEvent] = search_event(event_type)
        if not event_class:
            self.logger.error(
                "An event is not recognized! Please report with your log to help us diagnose."
            )
            raise ValueError(f"Unable to find event: {event_type}", data)
        data = {k: v for k, v in data.items() if k != "type"}
        event = event_class.parse_obj(data)
        return event

    async def ws_send_handle(
            self,
            command_name: str, data: Optional[dict] = None, subcommand: Optional[str] = None):
        if not self.bot_session.verifyKey:
            raise ValueError
        sync_id = str(random.randint(0, 100_000_000))
        content = {
            'syncId': sync_id,
            'main': command_name,
            'content': json.dumps(data)
        }
        if subcommand:
            content['subcommand'] = subcommand
        await self.ws_connection.send_json(content)
        future = self.loop.create_future()
        self.wait_response_future[sync_id] = future
        result = await future
        del self.wait_response_future[sync_id]
        del future
        if "data" in result:
            return result['data']
        else:
            return result

    async def send_handle(
            self,
            action: str,
            method: str,
            data: Optional[dict] = None
    ):
        if not self.bot_session.verifyKey:
            raise ValueError
        data = data or dict()
        if method == "GET" or method == "get":
            async with self.client_session.get(
                    URL(f"{self.bot_session.host}/{action}").with_query(data)
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        elif method == "POST" or method == "update":
            async with self.client_session.post(
                    URL(f"{self.bot_session.host}/{action}"), data=json.dumps(data)
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        else:
            form = aiohttp.FormData()
            for k, v in data:
                form.add_fields(k, v)
            async with self.client_session.post(
                    URL(f"{self.bot_session.host}/{action}"), data=form
            ) as response:
                response.raise_for_status()
                response_data = await response.json()
        if "data" in response_data:
            resp = response_data['data']
        else:
            resp = response_data
        error_check(response_data)
        return resp

    async def ws_receive_handle(self, unknown_event_data: dict):
        if "syncId" in unknown_event_data:
            data, sync_id = unknown_event_data.get("data"), unknown_event_data.get("syncId")
            if sync_id == "-1":
                event = await self.parse_to_event(data)
                with enter_context(bot=self.bot, event_i=event):
                    self.event_system.event_spread(event)
            else:
                if sync_id not in self.wait_response_future:
                    self.logger.warning(f"syncId {sync_id} not found!")
                else:
                    self.wait_response_future.pop(sync_id).set_result(data)

    async def receive_handle(self, unknown_event_data: dict):
        received_data = unknown_event_data.get('data')
        error_check(received_data)
        event = await self.parse_to_event(received_data)
        with enter_context(bot=self.bot, event_i=event):
            self.event_system.event_spread(event)

    async def websocket(self):
        query = {"qq": self.bot_session.account, "verifyKey": self.bot_session.verifyKey}
        if self.bot_session.single_mode:
            del query['qq']
        async with self.client_session.ws_connect(
                str(URL(self.bot_session.host + "/all").with_query(query)),
                autoping=False,
        ) as connection:
            self.logger.debug("connecting to websocket")
            self.ws_connection = connection
            connected = False
            while self.running:
                try:
                    ws_message = await connection.receive(timeout=self.timeout)
                except asyncio.TimeoutError:
                    try:
                        try:
                            self.logger.debug("websocket: trying ping...")
                            await self.ws_connection.ping()
                        except Exception as e:
                            self.logger.exception(f"websocket: ping failed: {e!r}")
                        else:
                            continue
                    except asyncio.CancelledError:
                        self.logger.warning("websocket: cancelled, stop")
                        return await self.stop()
                if ws_message.type is WSMsgType.TEXT:
                    received_data: dict = json.loads(ws_message.data)
                    if connected:
                        try:
                            await self.receive_handle(received_data)
                        except Exception as e:
                            self.logger.exception(f"receive_data has error {e}")
                    else:
                        if not received_data['syncId']:
                            data = received_data['data']
                            if data['code']:
                                error_check(data)

                            else:
                                if not self.bot_session.sessionKey:
                                    self.bot_session.sessionKey = data.get("session")
                                    connected = True
                elif ws_message.type is WSMsgType.CLOSE:
                    self.logger.info("websocket: server close connection.")
                    return
                elif ws_message.type is WSMsgType.CLOSED:
                    self.logger.info("websocket: connection has been closed.")
                    return
                elif ws_message.type is WSMsgType.PONG:
                    self.logger.debug("websocket: received pong from remote")
                elif ws_message.type == WSMsgType.ERROR:
                    self.logger.warning("websocket: connection error: " + ws_message.data)
                else:
                    self.logger.warning(f"detected a unknown message type: {ws_message.type}")
        self.logger.info("connection disconnected")

    async def connect(self):
        if not self.client_session:
            self.client_session = ClientSession()
        if not self.running_task or self.running_task.done():
            self.running = True
            self.running_task = self.loop.create_task(self.websocket())



from datetime import datetime
from typing import Optional, Literal

from pydantic import Field

from .base import MiraiEvent
from ..model.relation import Member, Group, Permission, Friend, Client
from .inserter import ApplicationInserter, EventInserter
from ..utils import bot_application, Structured
from ..message.messageChain import MessageChain


class BotEvent(MiraiEvent):
    """
    指示有关 Bot 本身的事件.
    """


class FriendEvent(MiraiEvent):
    """
    指示有关好友的事件.
    """


class GroupEvent(MiraiEvent):
    """
    指示有关群组的事件.
    """


class BotOnlineEvent(BotEvent):
    """
    Bot登录成功

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
    """

    type = "BotOnlineEvent"
    qq: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter
        )


class BotOfflineEventActive(BotEvent):
    """Bot主动离线

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
    """

    type = "BotOfflineEventActive"
    qq: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter
        )


class BotOfflineEventForce(BotEvent):
    """Bot被挤下线

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
    """

    type = "BotOfflineEventForce"
    qq: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter
        )


class BotOfflineEventDropped(BotEvent):
    """Bot被服务器断开或因网络问题而掉线

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
    """

    type = "BotOfflineEventDropped"
    qq: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter
        )


class BotReloginEvent(BotEvent):
    """Bot主动重新登录

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
    """

    type = "BotReloginEvent"
    qq: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter
        )


class FriendInputStatusChangedEvent(FriendEvent):
    """好友输入状态改变;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Friend (annotation): 发生该事件的好友
    """
    type = "FriendInputStatusChangedEvent"
    friend: Friend
    inputting: bool

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Friend=self.friend
        )


class FriendNickChangedEvent(FriendEvent):
    """好友昵称改变;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Friend (annotation): 发生该事件的好友
    """
    type = "FriendNickChangedEvent"
    friend: Friend
    name_from: str = Field(..., alias="from")
    name_to: str = Field(..., alias="to")

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Friend=self.friend
        )


class BotGroupPermissionChangeEvent(GroupEvent):
    """Bot在群里的权限被改变. 操作人一定是群主;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
    """

    type = "BotGroupPermissionChangeEvent"
    origin: Permission
    current: Permission
    group: Group

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Group=self.group
        )


class BotMuteEvent(GroupEvent):
    """Bot被禁言;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Member (annotation): 执行禁言操作的管理员/群主
        Group (annotation): 发生该事件的群组
    """

    type = "BotMuteEvent"
    durationSeconds: int
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.operator.group
        )


class BotUnmuteEvent(GroupEvent):
    """Bot被取消禁言;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Member (annotation): 执行解除禁言操作的管理员/群主, 若为 None 则为应用实例所辖账号操作
        Group (annotation): 发生该事件的群组
    """

    type = "BotUnmuteEvent"
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.operator.group
        )


class BotJoinGroupEvent(GroupEvent):
    """Bot加入了一个新群

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Group (annotation): 发生该事件的群组
    """

    type = "BotJoinGroupEvent"
    group: Group

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Group=self.group
        )


class BotLeaveEventActive(GroupEvent):
    """Bot主动退出一个群

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Group (annotation): 发生该事件的群组
    """

    type: str = "BotLeaveEventActive"
    group: Group

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Group=self.group
        )


class BotLeaveEventKick(GroupEvent):
    """Bot被踢出一个群

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Member (annotation): 执行本操作的群成员
        Group (annotation): 发生该事件的群组
    """

    type: str = "BotLeaveEventKick"
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Group=self.group,
            Member=self.operator
        )


class GroupRecallEvent(GroupEvent):
    """群消息被撤回;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Member (annotation): 执行本操作的群成员
        Group (annotation): 发生该事件的群组
    """

    type = "GroupRecallEvent"
    authorId: int
    messageId: int
    time: datetime
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class FriendRecallEvent(FriendEvent):
    """好友消息被撤回;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
    """

    type = "FriendRecallEvent"
    authorId: int
    messageId: int
    time: int
    operator: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )


class NudgeSubject(Structured):
    id: int
    kind: str


class NudgeEvent(GroupEvent):
    """戳一戳事件;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
    """

    type: str = "NudgeEvent"
    fromId: int
    action: str
    suffix: str
    target: int
    subject: NudgeSubject

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )


class GroupNameChangeEvent(GroupEvent):
    """某个群名被改变;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 被修改了群名称的群组
        Member (annotation): 更改群名称的成员, 权限必定为管理员或是群主
    """

    type = "GroupNameChangeEvent"
    origin: str
    current: str
    group: Group
    operator: Optional[Member] = None

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class GroupEntranceAnnouncementChangeEvent(GroupEvent):
    """某群入群公告被改变;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 被修改了入群公告的群组
        Member (annotation): 作出此操作的管理员/群主
    """

    type = "GroupEntranceAnnouncementChangeEvent"
    origin: str
    current: str
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class GroupMuteAllEvent(GroupEvent):
    """有一群组开启/关闭了全体禁言;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 全体禁言状态改变的群组
        Member (annotation): 作出此操作的管理员/群主
    """

    type = "GroupMuteAllEvent"
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class GroupAllowAnonymousChatEvent(GroupEvent):
    """有一群组开启/关闭了匿名聊天;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 修改了相关设定的群组
        Member (annotation): 作出此操作的管理员/群主
    """

    type = "GroupAllowAnonymousChatEvent"
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class GroupAllowConfessTalkEvent(GroupEvent):
    """有一群组开启/关闭了坦白说;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 修改了相关设定的群组
    """

    type = "GroupAllowAnonymousChatEvent"
    origin: bool
    current: bool
    group: Group
    isByBot: bool

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Group=self.group
        )


class GroupAllowMemberInviteEvent(GroupEvent):
    """有一群组修改了该群是否允许群员邀请其他用户加入群组的功能;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 修改了该功能的群组
        Member (annotation): 作出此操作的管理员/群主
    """

    type = "GroupAllowMemberInviteEvent"
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.operator,
            Group=self.group
        )


class MemberJoinEvent(GroupEvent):
    """新人入群的事件

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Group (annotation): 该用户加入的群组
        Member (annotation): 该用户的信息
    """

    type = "MemberJoinEvent"
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Member=self.member,
            Group=self.member.group
        )


class MemberLeaveEventKick(GroupEvent):
    """成员被踢出群（该成员不是Bot）;当 `operator` 为 `None` 时, 执行者为Bot

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Group (annotation): 指定的群组
        Member (annotation):
          - `"target"` (default, const, str): 被踢者的信息
          - `"operator"` (default, const, str): 执行了该操作的管理员/群主, 也可能是Bot.
    """

    type = "MemberLeaveEventKick"
    member: Member
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Group=self.member.group,
            target=self.member,
            operator=self.operator
        )


class MemberLeaveEventQuit(GroupEvent):
    """成员主动离群（该成员不是Bot）

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Group (annotation): 发生本事件的群组, 通常的, 在本事件发生后本群组成员数量少于之前
        Member (annotation): 退群群员的信息
    """

    type = "MemberLeaveEventQuit"
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Member=self.member,
            Group=self.member.group
        )


class MemberCardChangeEvent(GroupEvent):
    """群名片改动;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation): 名片改动的群员的信息
    """

    type = "MemberCardChangeEvent"
    origin: str
    current: str
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.member,
            Group=self.member.group
        )


class MemberSpecialTitleChangeEvent(GroupEvent):
    """群头衔改动（只有群主有操作限权）;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation): 被更改群头衔的群组成员
    """

    type = "MemberSpecialTitleChangeEvent"
    origin: str
    current: str
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.member,
            Group=self.member.group
        )


class MemberPermissionChangeEvent(GroupEvent):
    """成员权限改变的事件（该成员不是Bot）;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation): 被调整权限的群组成员
    """

    type = "MemberPermissionChangeEvent"
    origin: str
    current: str
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.member,
            Group=self.member.group
        )


class MemberMuteEvent(GroupEvent):
    """群成员被禁言事件（该成员不是Bot）;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation):
          - `"target"` (default, const, str): 被禁言的群员的信息
          - `"operator"` (default, const, str): 操作者的信息，当None时为Bot操作
    """

    type = "MemberMuteEvent"
    durationSeconds: int
    member: Member
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Group=self.member.group,
            target=self.member,
            operator=self.operator
        )


class MemberUnmuteEvent(GroupEvent):
    """群成员被取消禁言事件（该成员不是Bot）;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation):
          - `"target"` (default, const, str): 被禁言的群员的信息
          - `"operator"` (default, const, str): 操作者的信息，当None时为Bot操作
    """

    type = "MemberUnmuteEvent"
    member: Member
    operator: Optional[Member]

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Group=self.member.group,
            target=self.member,
            operator=self.operator
        )


class MemberHonorChangeEvent(GroupEvent):
    """群员称号改变;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Group (annotation): 发生该事件的群组
        Member (annotation): 称号改变的群组成员
    """
    type = "MemberHonorChangeEvent"
    action: str
    honor: Literal["achieve", "lose"]
    member: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Member=self.member,
            Group=self.member.group
        )


class RequestEvent(MiraiEvent):
    type: str
    eventId: int
    fromId: int
    groupId: Optional[int]
    nick: str
    message: Optional[str] = ""

    async def response_event(self, operate: int, msg: Optional[str] = ""):
        bot = bot_application.get()
        await bot.communicator.send_handle(
            f"resp/{self.type[0].lower() + self.type[1:]}",
            "POST",
            {
                "sessionKey": bot.bot_session.sessionKey,
                "eventId": self.eventId,
                "fromId": self.fromId,
                "groupId": self.groupId,
                "operate": operate,
                "message": msg
            }
        )


class NewFriendRequestEvent(RequestEvent):
    """添加好友申请;可使用原始事件类作为参数, 以此获得该事件实例, 以获取更多的信息

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例

    参数说明:
        1. event.fromId: int - 申请人QQ号
        2. event.groupId: int - 申请人如果通过某个群添加好友，该项为该群群号；否则为0
        3. event.nick: str  - 申请人的昵称或群名片
        4. event.message: str  - 申请消息
    方法说明:
        1. 同意添加好友: await event.accept()
        2. 拒绝添加好友: await event.reject()
        3. 拒绝并不再接收该用户的好友申请: await event.block()
    """
    type = "NewFriendRequestEvent"

    async def accept(self, message: str = ""):
        await self.response_event(0, message)

    async def reject(self, message: str = ""):
        await self.response_event(1, message)

    async def block(self, message: str = ""):
        await self.response_event(2, message)

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )


class MemberJoinRequestEvent(RequestEvent):
    """用户入群申请;该事件需要你将事件实例作为参数，以进行操作

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例

    参数说明:
        1. event.fromId: int - 申请人QQ号
        2. event.groupId: int - 申请人申请入群的群号
        3. event.groupName: str - 申请人申请入群的群名称
        4. event.nick: str  - 申请人的昵称或群名片
        5. event.message: str  - 申请消息
    方法说明:
        1. 同意入群: await event.accept()
        2. 拒绝入群: await event.reject()
        3. 忽略请求: await event.ignore()
        4. 拒绝并不再接收该用户的入群申请: await event.reject_and_block()
        5. 忽略并不再接收该用户的入群申请: await event.ignore_and_block()
    """
    type = "MemberJoinRequestEvent"
    groupName: str

    async def accept(self, message: str = ""):
        await self.response_event(0, message)

    async def reject(self, message: str = ""):
        await self.response_event(1, message)

    async def ignore(self, message: str = ""):
        await self.response_event(2, message)

    async def reject_and_block(self, message: str = ""):
        await self.response_event(3, message)

    async def ignore_and_block(self, message: str = ""):
        await self.response_event(4, message)

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )


class BotInvitedJoinGroupRequestEvent(RequestEvent):
    """Bot被邀请入群申请;该事件需要你将事件实例作为参数，以进行操作

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例

        Event (annotation): 该事件的实例

    参数说明:
        1. event.fromId: int - 邀请人（好友）的QQ号
        2. event.groupId: int - 被邀请进入群的群号
        3. event.groupName: str - 被邀请进入群的群名称
        4. event.nick: str  - 邀请人（好友）的昵称
        5. event.message: str  - 邀请消息
    方法说明:
        1. 同意邀请: await event.accept()
        2. 拒绝邀请: await event.reject()
    """
    type = "BotInvitedJoinGroupRequestEvent"
    groupName: str

    async def accept(self, message: str = ""):
        await self.response_event(0, message)

    async def reject(self, message: str = ""):
        await self.response_event(1, message)

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )


class OtherClientOnlineEvent(MiraiEvent):
    """其他客户端上线;该事件需要你将事件实例作为参数，以进行操作

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Event (annotation): 该事件的实例
        Client (annotation): 其他客户端的信息
    """
    type = "OtherClientOnlineEvent"
    client: Client
    kind: int

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            Client=self.client
        )


class OtherClientOfflineEvent(MiraiEvent):
    """其他客户端下线

    该事件可提供的参数:
        Cesloi (annotation): 发布事件的应用实例
        Client (annotation): 其他客户端的信息
    """
    type = "OtherClientOfflineEvent"
    client: Client

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            Client=self.client
        )


class CommandExecutedEvent(MiraiEvent):
    type = "CommandExecutedEvent"
    name: str
    args: MessageChain

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.args
        )



from datetime import datetime, timedelta
from typing import Dict, Union
import abc


class Timer(abc.ABC):
    type: str
    interval: Dict[str, Union[int, float]] = {}

    @abc.abstractmethod
    def get_delta(self):
        pass


class EveryTimer(Timer):
    """
    从机器人启动时计时
    """
    type: str = 'every'

    def __init__(self, **kwargs):
        if kwargs:
            self.interval = kwargs
        else:
            self.every_minute()

    def get_delta(self):
        while True:
            yield datetime.now() + timedelta(**self.interval)

    def every_second(self):
        """每秒执行一次
        """
        self.interval = {"seconds": 1}
        return self

    def every_custom_seconds(self, seconds: int):
        """每 seconds 秒执行一次

        Args:
            seconds (int): 距离下一次执行的时间间隔, 单位为秒
        """
        self.interval = {"seconds": seconds}
        return self

    def every_minute(self):
        """每分钟执行一次
        """
        self.interval = {'minutes': 1}
        return self

    def every_custom_minutes(self, minutes: int):
        """每 minutes 分钟执行一次

        Args:
            minutes (int): 距离下一次执行的时间间隔, 单位为分
        """
        self.interval = {"seconds": minutes}
        return self

    def every_hour(self):
        """每小时执行一次
        """
        self.interval = {'hours': 1}
        return self

    def every_custom_hours(self, hours: int):
        """每 hours 小时执行一次

        Args:
            hours (int): 距离下一次执行的时间间隔, 单位为小时
        """
        self.interval = {'hours': hours}
        return self

    def every_week(self):
        """每隔一周执行一次
        """
        self.interval = {'weeks': 1}
        return self

    def every_custom_weeks(self, weeks: int):
        """每 weeks 周执行一次

        Args:
            weeks (int): 距离下一次执行的时间间隔, 单位为周
        """
        self.interval = {'weeks': weeks}
        return self

    def every_day(self):
        """每隔一天执行一次
        """
        self.interval = {'days': 1}
        return self

    def every_custom_days(self, days: int):
        """每 days 天执行一次

        Args:
            days (int): 距离下一次执行的时间间隔, 单位为天
        """
        self.interval = {'days': days}
        return self


class RouteTimer(Timer):
    """
    从对应的0时刻开始计时
    """
    type: str = 'route'

    def get_delta(self):
        while True:
            yield self._start_time()

    def __init__(self, **kwargs):
        if kwargs:
            self.interval = kwargs

    def route_second(self, seconds: int):
        """在每分钟的第 seconds 秒执行一次

        Args:
            seconds (int): 距离下一次执行的时间间隔, 单位为秒
        """
        if seconds > 60 or seconds < 0:
            raise ValueError
        self.interval['seconds'] = seconds
        return self

    def route_minutes(self, minutes: int):
        """在每小时的第 minutes 分执行一次

        Args:
            minutes (int): 距离下一次执行的时间间隔, 单位为分
        """
        if minutes > 60 or minutes < 0:
            raise ValueError
        self.interval['minutes'] = minutes
        return self

    def route_hours(self, hours: int):
        """在每天的第 hours 时执行一次

        Args:
            hours (int): 距离下一次执行的时间间隔, 单位为时
        """
        if hours > 24 or hours < 0:
            raise ValueError
        self.interval['hours'] = hours
        return self

    def route_weekdays(self, weekday: int):
        """在每周的星期 weekday 执行一次

        Args:
            weekday (int): 星期几，从monday = 1开始算起
        """
        if weekday > 7 or weekday < 0:
            raise ValueError
        self.interval['weeks'] = weekday / 7
        return self

    def route_days(self, days: int):
        """在每月的 days 号执行一次

        Args:
            days (int): 距离下一次执行的时间间隔, 单位为天
        """
        if days > 31 or days < 0:
            raise ValueError
        self.interval['days'] = days
        return self

    def _start_time(self):
        now = datetime.now()
        year = now.year
        month = now.month
        weekday = now.weekday()
        day = now.day
        hour = now.hour
        minute = now.minute
        second = now.second
        for k, v in self.interval.items():
            if k == 'seconds':
                if v - second < 0:
                    minute += 1
                second = v
            if k == 'minutes':
                if v - minute < 0:
                    hour += 1
                minute = v
                second = 0
            if k == 'hours':
                if v - hour < 0:
                    day += 1
                hour = v
                minute = 0
                second = 0

            if k == 'weeks':
                hour = 0
                minute = 0
                second = 0
                day += 6 - weekday

            if k == 'days':
                if v - day < 0:
                    month += 1
                day = v
                hour = 0
                minute = 0
                second = 0
        return datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second)


class SpecialTimer(Timer):
    """
    直到一年中一个特定的时间才执行
    """

    def __init__(self, **kwargs):
        if kwargs:
            self.interval = kwargs

    def get_delta(self):
        while True:
            yield self._start_time()

    def set_special_day(
            self,
            month: int,
            day: int,
            hour: int = 0,
            minute: int = 0,
            second: int = 0
    ):
        self.interval = {'month': month, 'day': day, 'hour': hour, 'minute': minute, 'second': second}
        return self

    def _start_time(self):
        now = datetime.now()
        year = now.year
        if self.interval['month'] < now.month:
            year += 1
        return datetime(year=year, **self.interval)


if __name__ == "__main__":
    import asyncio

    a = RouteTimer().route_second(7)
    print(a._start_time() + timedelta(**a.interval))
    c = (a._start_time() + timedelta(**a.interval) - datetime.now()).total_seconds()
    print(c)
    asyncio.run(asyncio.sleep(c))
    print(datetime.now())

    b = EveryTimer().every_custom_seconds(20)
    d = (datetime.now() + timedelta(**b.interval) - datetime.now()).total_seconds()
    print(d)
    asyncio.run(asyncio.sleep(d))
    print(datetime.now())

    c = SpecialTimer().set_special_day(12, 1, 12)
    print(c._start_time() - datetime.now())



from arclet.letoderea.entities.event import Insertable
from arclet.cesloi.utils import bot_application, event


class ApplicationInserter(Insertable):
    name: str = "bot_application"

    @classmethod
    def get_params(cls):
        return cls.param_export(
            Cesloi=bot_application.get()
        )


class EventInserter(Insertable):
    name: str = "event"

    @classmethod
    def get_params(cls):
        return cls.param_export(
            event=event.get()
        )


if __name__ == "__main__":
    print(EventInserter.get_params())



from typing import List, Union, Tuple, Optional
from enum import Enum
import re


class Argument(Enum):
    AnyStr = "@A@"
    Album = "@B@"
    Digit = "@C@"


class Subcommand:
    name: Union[str, List[str]]
    separate: str
    args: Union[str, "Subcommand", List["Subcommand"]]

    def __init__(self, name, args: Optional[Union[str, "Subcommand", List["Subcommand"]]] = None, separate=" "):
        self.name = name
        self.args = args or ""
        self.separate = "" if self.args == "" else separate
        self.name = self.name.replace("@A@", r"(.+)").replace("@C@", r"(\d+)").replace("@B@", r"(\w+)")
        if isinstance(self.args, str):
            self.args = self.args.replace("@A@", "(.+)").replace("@C@", r"(\d+)").replace("@B@", r"(\w+)")

    def analysis_content(self):
        if isinstance(self.args, str):
            return [self.name + self.separate + self.args]
        elif isinstance(self.args, Subcommand):
            if self.name.rstrip(' ') == "":
                return [self.name + self.separate + sub for sub in self.args.analysis_content()]
            return [self.name + self.separate + sub for sub in self.args.analysis_content()] + [self.name]
        elif isinstance(self.args, list):
            result = []
            for sub in self.args:
                result.extend(sub.analysis_content())
            if self.name.rstrip(' ') == "":
                return [self.name + self.separate + sub for sub in result]
            return [self.name + self.separate + sub for sub in result] + [self.name]


class Command:
    """命令/命令参数解析器

    样例：Command(headers=[""], main=["name","argument/subcommand/subcommand_list","separate"])

    其中
        - name: 命令名称, 必写
        - argument: 命令参数, 可选
        - separate: 命令分隔符,分隔name与args,通常情况下为 " " (空格), 可选
        - subcommand: 子命令, 格式与main相同
        - subcommand_list: 子命令集, 可传入多个子命令
    name与args接受Command提供的三个参数: AnyStr, Album, Digit, 即name/args接受 任意字符/字母数字组合/纯数字

    您也可以用自己的写的正则表达式,前提是需要用括号括起来,如(.*?)

    Args:
        headers: 呼叫该命令的命令头，一般是你的机器人的名字，可选
        main: 命令主体，你的命令的主要参数解析部分，可选
    """
    headers: Optional[List[str]]
    main: Optional[Subcommand]

    def __init__(self, headers: Optional[List[str]] = None, main: Optional[list] = None):
        if not headers and not main:
            raise ValueError("Must choose one parameter!")
        self.headers = headers or [""]
        if main:
            self.main = self.to_subcommand(main)
        else:
            self.main = None

    def to_subcommand(self, cl):
        if not cl:
            raise ValueError("Must input name!")
        name = cl[0]
        if len(cl) > 1:
            args = cl[1]
            if isinstance(args, list):
                if isinstance(args[0], str):
                    args = self.to_subcommand(args)
                else:
                    args = []
                    for sub in cl[1]:
                        args.append(self.to_subcommand(sub))
            if len(cl) > 2:
                sep = cl[2]
                return Subcommand(name, args, separate=sep)
            return Subcommand(name, args)
        return Subcommand(name)

    def analysis(self, cmd: str) -> Union[str, Tuple[str], bool]:
        cmd = cmd.rstrip(' ')
        for head in self.headers:
            if head != "" and head in cmd:
                cmd = cmd.replace(head, "", 1)
                break
        cmd = cmd.lstrip(' ')
        if not self.main:
            if cmd == "":
                return True
            return False
        for pat in self.main.analysis_content():
            pattern = re.compile('^' + pat + '$')
            result = pattern.findall(cmd)
            if result:
                if result[0] == cmd:
                    return True
                return result[0]
        return False


AnyStr = Argument.AnyStr.value
Album = Argument.Album.value
Digit = Argument.Digit.value

if __name__ == "__main__":
    """
    演示程序
    """
    v = Command(main=["img", [["download", ["-p", AnyStr]],
                              ["upload", [["-u", AnyStr], ["-f", AnyStr]]]]])
    print(v.analysis("img upload -u http://www.baidu.com"))  # http://www.baidu.com
    print(v.analysis("img upload -f img.png"))  # img.png

    v = Command(headers=['bot', 'cmd.'], main=["", [["签到"], ["sign in"]], ""])
    print(v.analysis("cmd.sign in"))  # True

    v = Command(headers=["bots", "bot"])
    print(v.analysis("bot"))  # True
    print(v.analysis("bots aaa"))  # False

    v = Command(headers=["bots", "bot"], main=[AnyStr])
    print(v.analysis("bot"))  # False
    print(v.analysis("bots aaa"))  # True

    v = Command(main=[".command", ["--option", "foo"]])
    print(v.analysis(".command --option foo"))  # True
    print(v.analysis(".command "))

    v = Command(main=[f"/ping {AnyStr}", [["-t", Digit], ["-a"]]])
    print(v.analysis("/ping 127.0.0.1 -t 10"))
    print(v.analysis("/ping 127.0.0.1"))
    print(v.analysis("/ping 127.0.0.1 -a"))



def split_once(text: str, separate: str):  # 相当于另类的pop
    out_text = ""
    quotation_stack = []
    is_split = True
    for char in text:
        if char in "'\"":  # 遇到引号括起来的部分跳过分隔
            if not quotation_stack:
                is_split = False
                quotation_stack.append(char)
            else:
                is_split = True
                quotation_stack.pop(-1)
        if separate == char and is_split:
            break
        out_text += char
    return out_text, text.replace(out_text, "", 1).replace(separate, "", 1)


def split(text: str, separate: str = " ", max_split: int = -1):
    text_list = []
    quotation_stack = []
    is_split = True
    while all([text, max_split]):
        out_text = ""
        for char in text:
            if char in "'\"":  # 遇到引号括起来的部分跳过分隔
                if not quotation_stack:
                    is_split = False
                    quotation_stack.append(char)
                else:
                    is_split = True
                    quotation_stack.pop(-1)
            if separate == char and is_split:
                break
            out_text += char
        text_list.append(out_text)
        text = text.replace(out_text, "", 1).replace(separate, "", 1)
        max_split -= 1
    if text:
        text_list.append(text)
    return text_list


from arclet.letoderea.entities.event import TemplateEvent
from pydantic import validator
from ..utils import Structured

from .inserter import ApplicationInserter, EventInserter


class MiraiEvent(Structured, TemplateEvent):
    type: str

    @classmethod
    @validator("type", allow_reuse=True)
    def type_limit(cls, v):
        if cls.type != v:
            raise Exception("{0}'s type must be '{1}', not '{2}'".format(cls.__name__, cls.type, v))
        return v

    class Config:
        extra = "ignore"


class EmptyEvent(MiraiEvent):

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter
        )



from ...letoderea.entities.event import TemplateEvent, ParamRet


class ScheduleTaskEvent(TemplateEvent):

    @classmethod
    def get_params(cls) -> ParamRet:
        return cls.param_export(
        )



from typing import TYPE_CHECKING
if not TYPE_CHECKING:
    import arclet.cesloi.event.inserter
    import arclet.cesloi.event.mirai
    import arclet.cesloi.event.messages
    import arclet.cesloi.event.lifecycle
    import arclet.cesloi.event.base



import os
import sys

import loguru
import time


info_format = (
        '<green>{time:YYYY-MM-DD HH:mm:ss.SS}</green> | <level>{level: <8}</level> | '
        '<cyan>{name}</cyan> - <level>{message}</level>'
    )
debug_format = (
        '<green>{time:YYYY-MM-DD HH:mm:ss.SSSS}</green> | <level>{level: <9}</level> | '
        '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level> '
    )


class Logger:
    logger: loguru.logger = loguru.logger

    def __init__(self, level='INFO'):
        LOG_FORMAT = debug_format if level == 'DEBUG' else info_format
        self.logger.remove()
        self.logger.add(
            os.path.join(os.getcwd(), 'logs', 'bot_running_log_' + time.strftime("%Y-%m-%d", time.localtime()) + '.log'),
            format=LOG_FORMAT,
            level=level,
            enqueue=True,
            rotation="00:00",
            compression='zip',
            encoding="utf-8",
            backtrace=True,
            diagnose=True,
            colorize=False,
        )
        self.logger.add(sys.stderr, level=level, format=LOG_FORMAT, backtrace=True, diagnose=True, colorize=True,)
        self.logger.info("--------------------------------------------------------------------")



import os
import sys
from types import ModuleType
from typing import Optional, Dict, Union, Type, Callable, List
import importlib
from ..letoderea import EventSystem, TemplateEvent, EventDelegate, Publisher, Subscriber, Condition_T, \
    TemplateDecorator, search_event, event_class_generator
from .logger import Logger


class TemplatePlugin:
    module: ModuleType
    name: str
    usage: str
    is_close: bool = False

    def __init__(self, module, name: Optional[str] = None, usage: Optional[str] = None):
        if hasattr(module.__init__, "__annotations__"):
            module.__init__(**module.__init__.__annotations__)
        self.module = module
        self.name = name or ""
        self.usage = usage or ""


class Bellidin:
    """
    贝利丁(Bellidin), Cesloi的哥哥

    Cesloi的插件管理器
     - set_bellidin: 初始化管理器，通常不需要管
     - install_plugin: 载入单个模块，需要提供相对路径
     - install_plugins: 载入文件夹下的所有模块，需要提供相对路径
    """
    ignore = ["__init__.py", "__pycache__"]
    _modules: Dict[str, "TemplatePlugin"] = {}
    _module_target_dict: Dict[str, list] = {}
    current_module_name: str
    event_system: EventSystem
    logger: Logger.logger
    plugins_dir: str

    @classmethod
    def model_register(
            cls,
            event: Union[str, Type[TemplateEvent]],
            *,
            priority: int = 16,
            conditions: List[Condition_T] = None,
            decorators: List[TemplateDecorator] = None,
    ):
        if not cls.event_system:
            raise RuntimeError("Delegate didn't existed!")
        if isinstance(event, str):
            name = event
            event = search_event(event)
            if not event:
                raise Exception(name + " cannot found!")

        events = [event]
        events.extend(event_class_generator(event))
        conditions = conditions or []
        decorators = decorators or []

        def register_wrapper(func: Callable):
            subscriber = Subscriber(
                callable_target=func,
                decorators=decorators
            )
            for e in events:
                may_publishers = cls.event_system.get_publisher(e)
                _event_handler = EventDelegate(event=e)
                _event_handler += subscriber
                if not may_publishers:
                    cls.event_system.publisher_list.append(Publisher(priority, conditions, _event_handler))
                else:
                    for m_publisher in may_publishers:
                        if m_publisher.equal_conditions(conditions):
                            m_publisher += _event_handler
                            break
                    else:
                        cls.event_system.publisher_list.append(Publisher(priority, conditions, _event_handler))
            if cls.current_module_name not in cls._module_target_dict:
                cls._module_target_dict[cls.current_module_name] = [[event, subscriber]]
            else:
                cls._module_target_dict[cls.current_module_name].append([event, subscriber])
            return func

        return register_wrapper

    @classmethod
    def _uninstall_subscriber(cls, module_name):
        if cls._module_target_dict.get(module_name):
            targets = cls._module_target_dict[module_name]
            for event_type, subscriber in targets:
                t_publishers = cls.event_system.get_publisher(event_type)
                for pub in t_publishers:
                    pub.internal_delegate[event_type.__name__].subscribers.remove(subscriber)
                    if not pub.internal_delegate[event_type.__name__].subscribers:
                        pub.remove_delegate(event_type)
                        if not pub.internal_delegate:
                            cls.event_system.remove_publisher(pub)
            del cls._module_target_dict[module_name]

    @classmethod
    def set_bellidin(cls, event_system, logger):
        cls.event_system = event_system
        cls.logger = logger
        cls.plugins_dir = ""
        return cls

    @classmethod
    def get_delegate(cls):
        return cls.event_system

    @classmethod
    def install_plugin(cls, modules_name: str):
        try:
            cls.current_module_name = modules_name
            module = importlib.import_module(modules_name, modules_name)

            name = getattr(module, '__name__', None)
            usage = getattr(module, '__usage__', None)
            cls._modules[modules_name] = TemplatePlugin(module, name, usage)
            cls.logger.debug(f"plugin: {module.__name__} is installed")
            return True
        except Exception as e:
            cls.logger.exception(e)
            return False

    @classmethod
    def install_plugins(cls, plugin_dir: str):
        cls.plugins_dir = plugin_dir
        plugin_count = 0
        for module in os.listdir(plugin_dir):
            if module in cls.ignore:
                continue
            if os.path.isdir(module):
                cls.install_plugin(f"{plugin_dir.replace('/', '.')}.{module}")
                plugin_count += 1
            else:
                cls.install_plugin(f"{plugin_dir.replace('/', '.')}.{module.split('.')[0]}")
                plugin_count += 1
        cls.logger.info(f"{plugin_count} plugin have been installed")
        return plugin_count

    @classmethod
    def get_plugins(cls):
        return cls._modules

    @classmethod
    def uninstall_plugins(cls, *args, **kwargs):
        plugin_count = 0
        _names = list(cls._modules.keys())
        for file in os.listdir(cls.plugins_dir):
            if file == "__pycache__":
                for pyc in os.listdir(cls.plugins_dir + "/__pycache__"):
                    os.remove(cls.plugins_dir + "/__pycache__/" + pyc)
        for module_name in _names:
            cls.logger.debug(f"plugin: {module_name} uninstalling")
            cls._uninstall_subscriber(module_name)
            module = cls._modules[module_name].module
            if hasattr(module, "__end__"):
                try:
                    module.__end__(
                        *args, **module.__end__.__annotations__
                    )
                except Exception as e:
                    cls.logger.exception(e)
                    module.is_close = False
                else:
                    cls.logger.debug(f"plugin: {module.__name__} is uninstalled")
                    module.is_close = True
            del cls._modules[module_name]
            plugin_count += 1
            if sys.modules.get(module_name):
                del sys.modules[module_name]

        cls._modules.clear()
        cls.logger.info(f"{plugin_count} plugins have been uninstalled successfully")
        return plugin_count

    @classmethod
    def uninstall_plugin(cls, module_name: str, *args, **kwargs):
        try:
            if module_name in cls._modules:
                cls._uninstall_subscriber(module_name)
                module = cls._modules[module_name].module
                if hasattr(module, "__end__"):
                    try:
                        module.__end__(
                            *args, **module.__end__.__annotations__
                        )
                    except Exception as e:
                        cls.logger.exception(e)
                        module.is_close = False
                    else:
                        cls.logger.debug(f"plugin: {module.__name__} is uninstalled")
                        module.is_close = True
                del cls._modules[module_name]
                if sys.modules.get(module_name):
                    del sys.modules[module_name]
            else:
                raise ValueError(f"No such plugin named {module_name}")
        except Exception as e:
            cls.logger.debug(e)
            pass

    @classmethod
    def reload_plugins(cls, new_plugins_dir: Optional[str] = None):
        try:
            if new_plugins_dir:
                cls.uninstall_plugins(cls.plugins_dir)
                cls.plugins_dir = new_plugins_dir
                cls.logger.info(f"reload plugins at \"./{new_plugins_dir}\"")
                return cls.install_plugins(new_plugins_dir)
            else:
                cls.logger.info(f"reload plugins at \"./{cls.plugins_dir}\"")
                plugin_count = 0
                _names = list(cls._modules.keys())
                for module_name in _names:
                    cls._uninstall_subscriber(module_name)
                    cls.current_module_name = module_name
                    importlib.invalidate_caches()
                    cls._modules[module_name].module = importlib.reload(cls._modules[module_name].module)
                    cls.logger.debug(f"plugin: {cls._modules[module_name].module.__name__} is reloaded")
                    plugin_count += 1

                cls.logger.info(f"{plugin_count} plugins have been reload successfully")
                return plugin_count
        except Exception as e:
            cls.logger.exception(e)
            pass



from enum import Enum
from typing import Optional, Union
from ..utils import Structured
from pydantic import Field


class Permission(Enum):
    """描述群成员在群组中的权限"""

    Member = "MEMBER"  # 普通成员
    Administrator = "ADMINISTRATOR"  # 管理员
    Owner = "OWNER"  # 群主


class Equipment(Enum):
    """客户端的设备名称信息"""

    Mobile = "MOBILE"  # 移动端
    Windows = "WINDOWS"  # win电脑
    MacOS = "MACOS"  # mac电脑


class Friend(Structured):
    """好友的信息."""

    id: int
    nickname: str
    remark: Optional[str]


class Stranger(Structured):
    """描述 Tencent QQ 中的陌生人."""
    id: int
    nickname: str
    remark: str


class Group(Structured):
    id: int
    name: str
    accountPerm: Permission = Field(..., alias="permission")

    def avatar(self):
        return f'https://p.qlogo.cn/gh/{self.id}/{self.id}'


class Member(Structured):
    id: int
    name: str = Field(..., alias="memberName")
    permission: Permission
    specialTitle: Optional[str] = None
    joinTimestamp: Optional[int] = None
    lastSpeakTimestamp: Optional[int] = None
    muteTimeRemaining: Optional[int] = None
    group: Group

    def avatar(self):
        return f'https://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'


class GroupConfig(Structured):
    """描述群组各项功能的设置."""

    name: str
    announcement: str
    confessTalk: bool
    allowMemberInvite: bool
    autoApprove: bool
    anonymousChat: bool

    class Config:
        allow_mutation = True


class MemberInfo(Structured):
    """描述群组成员的各项功能的设置（需要有相关限权）."""

    id: int
    name: str = Field("", alias="memberName")
    specialTitle: str = ""
    permission: Permission
    joinTimestamp: Optional[int] = None
    lastSpeakTimestamp: Optional[int] = None
    muteTimeRemaining: Optional[int] = None
    group: Group

    class Config:
        allow_mutation = True

    def avatar(self):
        return f'https://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'


class Client(Structured):
    id: int
    platform: Equipment


Sender = Union[Friend, Member, Client, Stranger]



import json as JSON
from xml import sax
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, TYPE_CHECKING, Union, List
from base64 import b64decode, b64encode
from ..utils import Structured
import aiohttp
from pydantic import validator, Field
from abc import ABC

if TYPE_CHECKING:
    from .messageChain import MessageChain


class MessageElement(ABC, Structured):
    type: str

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    def to_serialization(self) -> str:
        return f"[mirai:{self.type}:{JSON.dumps(self.dict(exclude={'type'}))}]".replace('\n', '\\n').replace('\t',
                                                                                                             '\\t')

    @staticmethod
    def from_json(json: Dict):
        """将json转换为元素"""

    def to_text(self) -> str:
        return ""


class MediaElement(MessageElement):
    url: Optional[str] = None
    base64: Optional[str] = None

    async def get_bytes(self) -> bytes:
        if self.url and not self.base64:
            async with aiohttp.request("GET", self.url) as response:
                if response.status != 200:
                    raise ConnectionError(response.status, await response.text())
                data = await response.content.readexactly(response.content_length)
                self.base64 = str(b64encode(data), encoding='utf-8')
                return data
        if self.base64:
            return b64decode(self.base64)

    def to_sendable(self, path: Optional[Union[Path, str]] = None, data_bytes: Optional[bytes] = None):
        if sum([bool(self.url), bool(path), bool(self.base64)]) > 1:
            raise ValueError("Too many binary initializers!")
        if path:
            if isinstance(path, str):
                path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f"{path} is not exist!")
            self.base64 = str(b64encode(path.read_bytes()), encoding='utf-8')
        elif data_bytes:
            self.base64 = str(b64encode(data_bytes), encoding='utf-8')


class Source(MessageElement):
    """表示消息在一个特定聊天区域内的唯一标识"""
    type: str = "Source"
    id: int
    time: int

    @staticmethod
    def from_json(json: Dict) -> "Source":
        return Source.parse_obj(json)


class Quote(MessageElement):
    """表示消息中回复其他消息/用户的部分, 通常包含一个完整的消息链(`origin` 属性)"""
    type: str = "Quote"
    id: int
    groupId: int
    senderId: int
    targetId: int
    origin: "MessageChain"

    @validator("origin", pre=True, allow_reuse=True)
    def _(cls, v):
        from .messageChain import MessageChain
        return MessageChain(v)

    def to_serialization(self) -> str:
        return f"[mirai:Quote:{{\"id\":{self.id},\"origin\":{self.origin}}}]"

    @staticmethod
    def from_json(json: Dict) -> "Quote":
        return Quote.parse_obj(json)


class Plain(MessageElement):
    type: str = "Plain"
    text: str

    def __init__(self, text: str, **kwargs) -> None:
        """实例化一个 Plain 消息元素, 用于承载消息中的文字.

        Args:
            text (str): 元素所包含的文字
        """
        super().__init__(text=text, **kwargs)

    def to_text(self):
        return self.text.replace('\n', '\\n').replace('\t', '\\t')

    def to_serialization(self) -> str:
        return self.text.replace('\n', '\\n').replace('\t', '\\t')

    @staticmethod
    def from_json(json: Dict) -> "Plain":
        return Plain.parse_obj(json)


class At(MessageElement):
    """该消息元素用于承载消息中用于提醒/呼唤特定用户的部分."""

    type: str = "At"
    target: int
    display: Optional[str] = None

    def __init__(self, target: int, **kwargs) -> None:
        """实例化一个 At 消息元素, 用于承载消息中用于提醒/呼唤特定用户的部分.

        Args:
            target (int): 需要提醒/呼唤的特定用户的 QQ 号(或者说 id.)
        """
        super().__init__(target=target, **kwargs)

    def to_text(self) -> str:
        return f"@{str(self.display)}" if self.display else f"@{self.target}"

    @staticmethod
    def from_json(json: Dict) -> "At":
        return At.parse_obj(json)


class AtAll(MessageElement):
    """该消息元素用于群组中的管理员提醒群组中的所有成员"""
    type: str = "AtAll"

    def to_text(self) -> str:
        return "@全体成员"

    @staticmethod
    def from_json(json: Dict):
        return AtAll()


class Face(MessageElement):
    """表示消息中所附带的表情, 这些表情大多都是聊天工具内置的."""
    type: str = "Face"
    faceId: int
    name: Optional[str] = None

    def to_text(self) -> str:
        return f"[表情:{self.name}]" if self.name else f"[表情:{self.faceId}]"

    @staticmethod
    def from_json(json: Dict) -> "Face":
        return Face.parse_obj(json)


class Xml(MessageElement):
    type = "Xml"
    xml: str

    def __init__(self, xml, *_, **__) -> None:
        super().__init__(xml=xml)

    def to_text(self) -> str:
        return "[XML消息]"

    def get_xml(self):
        return sax.parseString(self.xml, sax.handler.ContentHandler())

    @staticmethod
    def from_json(json: Dict):
        return Xml.parse_obj(json)


class Json(MessageElement):
    type = "Json"
    Json: str = Field(..., alias="json")

    def __init__(self, json: Union[dict, str], **kwargs) -> None:
        if isinstance(json, dict):
            json = JSON.dumps(json)
        super().__init__(json=json, **kwargs)

    def to_text(self) -> str:
        return "[JSON消息]"

    def get_json(self) -> dict:
        return JSON.loads(self.Json)

    @staticmethod
    def from_json(json: Dict):
        return Json.parse_obj(json)


class App(MessageElement):
    type = "App"
    content: str

    def to_text(self) -> str:
        return f"[APP消息:{JSON.loads(self.content)['prompt']}]"

    def get_meta_content(self) -> dict:
        return JSON.loads(self.content)['meta']

    @staticmethod
    def from_json(json: Dict):
        return App.parse_obj(json)


class PokeMethods(Enum):
    ChuoYiChuo = "ChuoYiChuo"
    BiXin = "BiXin"
    DianZan = "DianZan"
    XinSui = "XinSui"
    LiuLiuLiu = "LiuLiuLiu"
    FangDaZhao = "FangDaZhao"
    BaoBeiQiu = "BaoBeiQiu"
    Rose = "Rose"
    ZhaoHuanShu = "ZhaoHuanShu"
    RangNiPi = "RangNiPi"
    JeiYin = "JeiYin"
    ShouLei = "ShouLei"
    GouYin = "GouYin"
    ZhuaYiXia = "ZhuaYiXia"
    SuiPing = "SuiPing"
    QiaoMen = "QiaoMen"


class Poke(MessageElement):
    type = "Poke"
    name: PokeMethods

    def to_text(self) -> str:
        return f"[戳一戳:{self.name}]"

    @staticmethod
    def from_json(json: Dict):
        return Poke.parse_obj(json)


class Dice(MessageElement):
    type = "Dice"
    value: int

    def to_text(self) -> str:
        return f"[骰子:{self.value}]"

    @staticmethod
    def from_json(json: Dict):
        return Dice.parse_obj(json)


class File(MessageElement):
    type = "File"
    file_id: str
    name: str
    size: int

    def to_text(self) -> str:
        return f'[文件:{self.name}]'

    @staticmethod
    def from_json(json: Dict):
        return File.parse_obj(json)


class ImageType(Enum):
    Friend = "Friend"
    Group = "Group"
    Temp = "Temp"
    Unknown = "Unknown"


class Image(MediaElement):
    """该消息元素用于承载消息中所附带的图片."""
    type = "Image"
    imageId: Optional[str] = None
    url: Optional[str] = None
    base64: Optional[str] = None

    def __init__(
            self,
            imageId: Optional[str] = None,
            url: Optional[str] = None,
            path: Optional[Union[Path, str]] = None,
            base64: Optional[str] = None,
            data_bytes: Optional[bytes] = None,
            **kwargs
    ):
        super().__init__(
            imageId=imageId,
            url=url,
            base64=base64,
            **kwargs
        )
        self.to_sendable(path, data_bytes)

    def to_text(self) -> str:
        return "[图片]"

    @staticmethod
    def from_json(json: Dict):
        return Image.parse_obj(json)


class FlashImage(Image):
    """该消息元素用于承载消息中所附带的图片."""
    type = "FlashImage"

    def __init__(
            self,
            imageId: Optional[str] = None,
            url: Optional[str] = None,
            path: Optional[Union[Path, str]] = None,
            base64: Optional[str] = None,
            data_bytes: Optional[bytes] = None,
            **kwargs
    ):
        super().__init__(
            imageId=imageId,
            url=url,
            base64=base64,
            **kwargs
        )
        self.to_sendable(path, data_bytes)

    def to_text(self) -> str:
        return "[闪照]"

    @staticmethod
    def from_json(json: Dict):
        return FlashImage.parse_obj(json)


class Voice(MediaElement):
    type = "Voice"
    voiceId: Optional[str]
    length: Optional[int]

    def __init__(
            self,
            voiceId: Optional[str] = None,
            url: Optional[str] = None,
            path: Optional[Union[Path, str]] = None,
            base64: Optional[str] = None,
            data_bytes: Optional[bytes] = None,
            **kwargs
    ):
        super().__init__(
            voiceId=voiceId,
            url=url,
            base64=base64,
            **kwargs
        )
        self.to_sendable(path, data_bytes)

    def to_text(self) -> str:
        return "[语音]"

    @staticmethod
    def from_json(json: Dict):
        return Voice.parse_obj(json)


class MusicShare(MessageElement):
    type = "MusicShare"
    kind: Optional[str]
    title: Optional[str]
    summary: Optional[str]
    jumpUrl: Optional[str]
    pictureUrl: Optional[str]
    musicUrl: Optional[str]
    brief: Optional[str]

    def to_text(self) -> str:
        return f"[音乐分享:{self.title}]"

    @staticmethod
    def from_json(json: Dict):
        return MusicShare.parse_obj(json)


class ForwardNode(Structured):
    """表示合并转发中的一个节点"""
    senderId: int
    time: int
    senderName: str
    messageChain: Optional["MessageChain"]
    messageId: Optional[int]


class Forward(MessageElement):
    """
    指示合并转发信息

    nodeList (List[ForwardNode]): 转发的消息节点
    """

    type = "Forward"
    nodeList: List[ForwardNode]

    def to_text(self) -> str:
        return f"[合并转发:共{len(self.nodeList)}条]"

    @staticmethod
    def from_json(json: Dict):
        return Forward.parse_obj(json)


def _update_forward_refs():
    from .messageChain import MessageChain

    Quote.update_forward_refs(MessageChain=MessageChain)
    ForwardNode.update_forward_refs(MessageChain=MessageChain)



import asyncio
import sys
import time
import traceback
from asyncio import Task
from typing import Optional, Union, List, Type

from arclet.cesloi.utils import enter_message_send_context, UploadMethods, bot_application_context_manager, \
    upload_method
from arclet.letoderea import EventSystem, Condition_T, TemplateDecorator, TemplateEvent
from arclet.cesloi.event.lifecycle import ApplicationRunning, ApplicationStop
from arclet.cesloi.event.messages import Message, GroupMessage, FriendMessage, TempMessage
from arclet.cesloi.logger import Logger
from arclet.cesloi.communicate_with_mah import BotSession, Communicator
from arclet.cesloi.model.utils import BotMessage, Profile, FileInfo
from arclet.cesloi.message.element import Source, MessageElement
from arclet.cesloi.model.relation import Group, Member, GroupConfig, MemberInfo, Friend
from arclet.cesloi.message.messageChain import MessageChain
from arclet.cesloi.plugin import Bellidin


class Cesloi:
    def __init__(
            self,
            *,
            bot_session: BotSession,
            event_system: Optional[EventSystem] = None,
            logger: Optional[Logger] = None,
            debug: bool = False,
            enable_chat_log: bool = True,
            use_loguru_traceback: Optional[bool] = True,
    ):
        self.event_system: EventSystem = event_system or EventSystem()
        self.bot_session: BotSession = bot_session
        self.debug = debug
        self.logger = logger or Logger(level='DEBUG' if debug else 'INFO').logger
        self.bellidin = Bellidin.set_bellidin(self.event_system, self.logger)
        self.chat_log_enabled = enable_chat_log
        self.communicator = Communicator(bot_session, bot=self, event_system=self.event_system, logger=self.logger)
        self.running: bool = False
        self.daemon_task: Optional[Task] = None
        self.group_message_log_format: str = "{bot_id}: [{group_name}({group_id})] {member_name}({member_id}) -> {" \
                                             "message_string} "
        self.friend_message_log_format: str = "{bot_id}: [{friend_name}({friend_id})] -> {message_string}"
        self.temp_message_log_format: str = "{bot_id}: [{group_name}({group_id}.{member_name}({member_id})] -> {" \
                                            "message_string} "

        if self.chat_log_enabled:
            self.event_system.register("GroupMessage")(self.logger_group_message)
            self.event_system.register("FriendMessage")(self.logger_friend_message)
            self.event_system.register("TempMessage")(self.logger_temp_message)

        if use_loguru_traceback:
            traceback.print_exception = self.loguru_excepthook
            sys.excepthook = self.loguru_excepthook
            self.event_system.loop.set_exception_handler(self.loguru_async_handler)

    @staticmethod
    def loguru_excepthook(cls, val, tb, *_, **__):
        Logger.logger.opt(exception=(cls, val, tb)).error(f"Exception:")

    @staticmethod
    def loguru_async_handler(loop, ctx: dict):
        Logger.logger.opt(exception=(Exception, ctx["message"], ctx["source_traceback"])).error(
            f"Exception:"
        )

    async def logger_group_message(self, event: GroupMessage):
        self.logger.info(
            self.group_message_log_format.format_map(
                dict(
                    group_id=event.sender.group.id,
                    group_name=event.sender.group.name,
                    member_id=event.sender.id,
                    member_name=event.sender.name,
                    bot_id=self.bot_session.account,
                    message_string=event.messageChain.__repr__(),
                )
            )
        )

    async def logger_friend_message(self, event: FriendMessage):
        self.logger.info(
            self.friend_message_log_format.format_map(
                dict(
                    bot_id=self.bot_session.account,
                    friend_name=event.sender.nickname,
                    friend_id=event.sender.id,
                    message_string=event.messageChain.__repr__(),
                )
            )
        )

    async def logger_temp_message(self, event: TempMessage):
        self.logger.info(
            self.temp_message_log_format.format_map(
                dict(
                    group_id=event.sender.group.id,
                    group_name=event.sender.group.name,
                    member_id=event.sender.id,
                    member_name=event.sender.name,
                    bot_id=self.bot_session.account,
                    message_string=event.messageChain.__repr__(),
                )
            )
        )

    async def running_task(self, retry_interval: float = 5.0):
        self.logger.debug("Cesloi Network Started.")
        while self.running:
            try:
                await self.communicator.connect()
                try:
                    if self.communicator.running_task:
                        await self.communicator.running_task
                except Exception as e:
                    self.logger.warning(e)
                await self.communicator.stop()
                self.logger.warning("Communicator stopped")
                await asyncio.sleep(retry_interval)
                self.logger.info("Cesloi Network Restarting...")
            except asyncio.CancelledError:
                await self.communicator.stop()
        self.logger.debug("Cesloi Network Stopped.")

    async def close(self):
        if self.running:
            self.event_system.event_spread(ApplicationStop(self))
            self.running = False
            self.uninstall_plugins()
            if self.daemon_task:
                self.daemon_task.cancel()
                self.daemon_task = None
            await self.communicator.stop()
            for t in asyncio.all_tasks(self.event_system.loop):
                if (
                        t is not asyncio.current_task(self.event_system.loop)
                        and not t.get_name().startswith("cesloi")
                        and not t.get_name().startswith("_")
                ):
                    t.cancel()
                    try:
                        await t
                    except asyncio.CancelledError:
                        pass

    def start(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        if self.event_system:
            loop = self.event_system.loop
        else:
            loop = loop or asyncio.get_event_loop()

        try:
            if not self.running:
                self.running = True
                start_time = time.time()
                self.logger.info("Cesloi Application Starting...")
                self.daemon_task = loop.create_task(self.running_task(), name="cesloi_web_task")
                while not self.bot_session.sessionKey:
                    loop.run_until_complete(asyncio.sleep(0.001))
                self.event_system.event_spread(ApplicationRunning(self))
                self.logger.info(f"Cesloi Application Started with {time.time() - start_time:.2}s")

            if self.daemon_task:
                loop.run_until_complete(self.daemon_task)
        except KeyboardInterrupt or asyncio.CancelledError:
            self.logger.warning("Interrupt detected, bot stopping ...")
        loop.run_until_complete(self.close())
        self.logger.info("Cesloi shutdown. Have a nice day!")

    def register(
            self, event: Union[str, Type[TemplateEvent]],
            *,
            priority: int = 16,
            conditions: List[Condition_T] = None,
            decorators: List[TemplateDecorator] = None
    ):
        """
        注册事件方法，用于指定订阅器订阅的发布器绑定的事件。
        """
        return self.event_system.register(event, priority=priority, conditions=conditions, decorators=decorators)

    def install_plugins(self, plugins_dir: str):
        """
        导入插件方法，必须传入一个目录，目录内有你的.py文件或模块
        """
        return self.bellidin.install_plugins(plugins_dir)

    def uninstall_plugins(self):
        """
        卸载导入的插件的方法.
        """
        return self.bellidin.uninstall_plugins()

    def reload_plugins(self, new_plugins_dir: Optional[str] = None):
        """
        重载插件方法，可选传入一个新目录，目录内有你的.py文件或模块
        """
        return self.bellidin.reload_plugins(new_plugins_dir)

    async def get_mah_version(self):
        result = await self.communicator.send_handle("about", "GET")
        return result['version']

    async def get_from_messageId(self, messageId: int):
        result = await self.communicator.send_handle(
            "messageFromId",
            "GET",
            {"sessionKey": self.bot_session.sessionKey, "id": messageId}
        )
        return MessageChain.parse_obj(result['messageChain'])

    async def get_friend_list(self):
        result = await self.communicator.send_handle(
            "friendList",
            "GET",
            {"sessionKey": self.bot_session.sessionKey}
        )
        return [Friend.parse_obj(i) for i in result]

    async def get_group_list(self):
        result = await self.communicator.send_handle(
            "groupList",
            "GET",
            {"sessionKey": self.bot_session.sessionKey}
        )
        return [Group.parse_obj(i) for i in result]

    async def get_member_list(self, group: Union[Group, int]):
        result = await self.communicator.send_handle(
            "memberList",
            "GET",
            {"sessionKey": self.bot_session.sessionKey,
             "target": group if isinstance(group, int) else group.id}
        )
        return [Member.parse_obj(i) for i in result]

    async def find(
            self,
            target_id: Optional[int] = None,
            group: Optional[Union[Group, int]] = None
    ) -> Optional[Union[Friend, Group, Member]]:
        """根据提供的参数自动尝试从好友列表、群列表或群成员列表中获取目标信息；可能返回None

        只填target则尝试在好友列表中搜索；只填group则尝试在群列表中搜索；两者都填则尝试在指定群组中搜索成员

        Args:
            target_id : 已知的好友id或者群成员id
            group : 尝试获取的已知的群组
        """
        if target_id and not group:
            data_friend = await self.get_friend_list()
            for i in data_friend:
                if i.id == target_id:
                    return i
        if group and not target_id:
            if isinstance(group, Group):
                group = group.id
            data_group = await self.get_group_list()
            for i in data_group:
                if i.id == group:
                    return i
        if group and target_id:
            data_member = await self.get_member_list(group)
            for i in data_member:
                if i.id == target_id:
                    return i

    async def get_bot_profile(self):
        result = await self.communicator.send_handle(
            "botProfile",
            "GET",
            {"sessionKey": self.bot_session.sessionKey}
        )
        return Profile.parse_obj(result)

    async def get_friend_profile(self, target: Union[Friend, int]):
        result = await self.communicator.send_handle(
            "friendProfile",
            "GET",
            {"sessionKey": self.bot_session.sessionKey,
             "target": target if isinstance(target, int) else target.id}
        )
        return Profile.parse_obj(result)

    async def get_member_profile(self, target: Union[Member, int], group: Optional[Union[Group, int]] = None):
        group = group or (target.group if isinstance(target, Member) else None)
        if not group:
            raise ValueError("Missing argument: group")
        result = await self.communicator.send_handle(
            "memberProfile",
            "GET",
            {"sessionKey": self.bot_session.sessionKey,
             "target": group if isinstance(group, int) else group.id,
             "memberId": target if isinstance(target, int) else target.id
             }
        )
        return Profile.parse_obj(result)

    async def get_profile(self, target: Optional[Union[Member, Friend, int]] = None,
                          group: Optional[Union[Group, int]] = None):
        """根据提供的参数自动尝试获取对象的资料

        若不填参数，则考虑获取机器人自身资料；若只填入target并且填入的为数字，则优先考虑获取好友资料；若填入group，则target为必填

        Args:
            target : 已知的好友或者群成员
            group : 已知的群组
        """
        if target and group:
            if isinstance(target, Member) or isinstance(target, int):
                return await self.get_member_profile(target, group)
            return await self.get_friend_profile(target)
        if target and not group:
            if isinstance(target, Friend) or isinstance(target, int):
                return await self.get_friend_profile(target)
            return await self.get_member_profile(target)
        if not target and group:
            raise ValueError("Missing argument: target")
        return await self.get_bot_profile()

    @bot_application_context_manager
    async def send_friend_message(
            self,
            target: Union[Friend, int],
            message: Union[MessageChain, str],
            *,
            quote: Optional[Union[Source, int]] = None,
    ) -> BotMessage:
        """使用此方法向指定好友发送消息, 可以指定需要回复的消息.

        Args:
            target : 指定的好友
            message : 消息链
            quote : 需要回复的消息源, 默认为 None.

        Returns:
            BotMessage:含有一 `messageId` 属性, 可用于回复.
        """
        if isinstance(message, str):
            message = MessageChain.create(message)
        with enter_message_send_context(UploadMethods.Friend):
            target_id = target.id if isinstance(target, Friend) else target
            result = await self.communicator.send_handle(
                "sendFriendMessage",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": target_id,
                    "messageChain": message.dict()["__root__"],
                    **(
                        {"quote": quote.id if isinstance(quote, Source) else quote} if quote else {}
                    )
                }
            )
            self.logger.info(f"[BOT {self.bot_session.account}] Friend({target_id}) <- {message.to_text()}")
            return BotMessage.parse_obj({"messageId": result['messageId']})

    @bot_application_context_manager
    async def send_group_message(
            self,
            target: Union[Group, int],
            message: Union[MessageChain, str],
            *,
            quote: Optional[Union[Source, int]] = None,
    ) -> BotMessage:
        """使用此方法向指定群组发送消息, 可以指定需要回复的消息.

        Args:
            target : 指定的群组.
            message : 消息链
            quote : 需要回复的消息源, 默认为 None.

        Returns:
            BotMessage:含有一 `messageId` 属性, 可用于回复.
        """
        if isinstance(message, str):
            message = MessageChain.create(message)
        with enter_message_send_context(UploadMethods.Group):
            target_id = target.id if isinstance(target, Group) else target
            result = await self.communicator.send_handle(
                "sendGroupMessage",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": target_id,
                    "messageChain": message.dict()["__root__"],
                    **(
                        {"quote": quote.id if isinstance(quote, Source) else quote} if quote else {}
                    )
                }
            )
            self.logger.info(f"[BOT {self.bot_session.account}] Group({target_id}) <- {message.to_text()}")
            return BotMessage.parse_obj({"messageId": result['messageId']})

    @bot_application_context_manager
    async def send_temp_message(
            self,
            target: Union[Member, int],
            message: Union[MessageChain, str],
            group: Optional[Union[Group, int]] = None,
            *,
            quote: Optional[Union[Source, int]] = None,
    ) -> BotMessage:
        """使用此方法向指定群组中的特定组员发送消息, 可指定需要回复的消息.

        Args:
            target : 指定的群组成员.
            group : 指定的群组.
            message : 消息链.
            quote : 需要回复的消息源, 默认为 None.

        Returns:
            BotMessage:含有一 `messageId` 属性, 可用于回复.
        """
        if isinstance(message, str):
            message = MessageChain.create(message)
        group = target.group if (isinstance(target, Member) and not group) else group
        if not group:
            raise ValueError("Missing argument: group")
        with enter_message_send_context(UploadMethods.Temp):
            group_id = group.id if isinstance(group, Group) else group
            target_id = target.id if isinstance(target, Member) else target
            result = await self.communicator.send_handle(
                "sendTempMessage",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "group": group_id,
                    "qq": target_id,
                    "messageChain": message.dict()["__root__"],
                    **(
                        {"quote": quote.id if isinstance(quote, Source) else quote} if quote else {}
                    )
                }
            )
            self.logger.info(
                f"[BOT {self.bot_session.account}] Member({target_id}, in {group_id}) <- {message.to_text()}")
            return BotMessage.parse_obj({"messageId": result['messageId']})

    @bot_application_context_manager
    async def send_with(
            self,
            target: Union[Friend, Group, Member, Message],
            messages: Optional[Union[MessageChain, str, List[MessageElement], MessageElement]] = None,
            *,
            quote: Optional[Union[Source, int]] = None,
            nudge: bool = False
    ):
        """应用发送消息操作的统合方法

        Args:
            target : 指定的好友、群组、群组成员、甚至是事件实例
            messages : 消息链
            quote : 需要回复的消息源, 默认为 None.
            nudge : 是否发送戳一戳，默认为False.

        Returns:
            BotMessage:含有一 `messageId` 属性, 可用于回复.

        """
        if isinstance(target, (GroupMessage, FriendMessage, TempMessage)):
            target = target.sender
        else:
            target = target
        if nudge:
            if isinstance(target, Friend) or isinstance(target, Member):
                await self.send_nudge(target)
        if messages:
            data = {"message": messages, "target": target}
            if quote:
                data["quote"] = quote
            if isinstance(target, Friend):
                return await self.send_friend_message(**data)
            elif isinstance(target, Group):
                return await self.send_group_message(**data)
            elif isinstance(target, Member):
                return await self.send_temp_message(**data)
            else:
                raise ValueError(
                    f"target {target} unclear! Please take an instance of the dialog object type as a parameter"
                )

    async def send_nudge(self, target: Union[Friend, Member, int]):
        target_id = target if isinstance(target, int) else target.id
        await self.communicator.send_handle(
            "sendNudge",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": target_id,
                "subject": target.group.id if isinstance(target, Member) else target_id,
                "kind": "Group" if isinstance(target, Member) else "Friend"
            }
        )
        self.logger.info(
            f"[BOT {self.bot_session.account}] Member({target_id}) <- Nudge")

    async def recall_message(self, target: Union[Source, BotMessage, int]):
        """
        使用此方法撤回一指定消息
        """
        if isinstance(target, Source):
            target = target.id
        if isinstance(target, BotMessage):
            target = target.messageId

        await self.communicator.send_handle(
            "recall",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": target,
            }
        )

    async def delete_friend(self, target: Union[Friend, int]):
        """
        使用此方法删除指定好友
        """
        friend_id = target.id if isinstance(target, Friend) else target
        await self.communicator.send_handle(
            "deletaFriend",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": friend_id,
            }
        )

    async def mute(
            self,
            target: Optional[Union[Member, int]] = None,
            group: Optional[Union[Group, int]] = None,
            mute_time: int = 60, mute_all: bool = False
    ):
        """
        使用此方法指定群禁言指定群员或开启全体禁言（需要有相关限权）

        Args:
            target : 指定成员.
            group : 指定群组.
            mute_time : 禁言时长，单位为秒，最多30天，默认为60秒.
            mute_all : 是否开启全体禁言,要求群组参数为必填.
        """
        if target:
            if not group:
                if isinstance(target, Member):
                    group_id = target.group.id
                else:
                    raise ValueError("Missing Argument: group")
            else:
                group_id = group.id if isinstance(group, Group) else group
            await self.communicator.send_handle(
                "mute",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": group_id,
                    "memberId": target.id if isinstance(target, Member) else target,
                    "time": mute_time
                }
            )
        elif group and mute_all:
            await self.communicator.send_handle(
                "muteAll",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": group.id if isinstance(group, Group) else group
                }
            )

    async def unmute(self, target: Optional[Union[Member, int]] = None, group: Optional[Union[Group, int]] = None,
                     unmute_all: bool = False):
        """
        使用此方法指定群解除群成员禁言或解除全体禁言（需要有相关限权）

        Args:
            target : 指定成员;
            group : 指定群组;
            unmute_all : 是否解除全体禁言,要求群组参数为必填;
        """
        if target:
            if not group:
                if isinstance(target, Member):
                    group_id = target.group.id
                else:
                    raise ValueError("Missing Argument: group")
            else:
                group_id = group.id if isinstance(group, Group) else group
            await self.communicator.send_handle(
                "unmute",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": group_id,
                    "memberId": target.id if isinstance(target, Member) else target
                }
            )
        elif group and unmute_all:
            await self.communicator.send_handle(
                "unmuteAll",
                "POST",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "target": group.id if isinstance(group, Group) else group
                }
            )

    async def kick_member(self, target: Union[Member, int], group: Optional[Union[Group, int]] = None,
                          msg: Optional[str] = ""):
        """
        使用此方法移除指定群成员（需要有相关限权）
        """
        if not group:
            if isinstance(target, Member):
                group_id = target.group.id
            else:
                raise ValueError("Missing Argument: group")
        else:
            group_id = group.id if isinstance(group, Group) else group
        await self.communicator.send_handle(
            "kick",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group_id,
                "memberId": target.id if isinstance(target, Member) else target,
                "msg": msg
            }
        )

    async def self_quit(self, group: Union[Group, int]):
        """
        使用此方法使Bot退出群聊
        """
        await self.communicator.send_handle(
            "quit",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group
            }
        )

    async def set_essence(self, target: Union[Source, BotMessage, int]):
        """
        使用此方法撤回一指定消息
        """
        if isinstance(target, Source):
            target = target.id
        if isinstance(target, BotMessage):
            target = target.messageId

        await self.communicator.send_handle(
            "setEssence",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": target,
            }
        )

    async def get_group_config(self, group: Union[Group, int]):
        result = await self.communicator.send_handle(
            "groupConfig",
            "get",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group
            }
        )

        return GroupConfig.parse_obj(result)

    async def get_member_info(self, member: Union[Member, int], group: Optional[Union[Group, int]] = None):
        if not group and isinstance(member, int):
            raise ValueError("You must give a Member when there isn't Group")
        if isinstance(member, Member) and not group:
            group = member.group
        result = await self.communicator.send_handle(
            "memberInfo",
            "get",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group,
                "memberId": member.id if isinstance(member, Member) else member
            }
        )
        return MemberInfo.parse_obj(result)

    async def get_info(self, target: Union[Group, int, Member]):
        """根据提供的参数自动尝试获取对象的信息或配置

        若想获取群成员信息则target必须为Member

        Args:
            target : 已知的群组或群成员
        """
        if isinstance(target, Member):
            return await self.get_member_info(target)
        return await self.get_group_config(target)

    async def set_group_config(self, group: Union[Group, int], config: GroupConfig):
        await self.communicator.send_handle(
            "groupConfig",
            "update",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group,
                "config": config.dict(exclude_unset=True, exclude_none=True)
            }
        )

    async def set_member_info(self, member: Union[Member, int], member_info: MemberInfo,
                              group: Optional[Union[Group, int]] = None):
        if not group and isinstance(member, int):
            raise ValueError("You must give a Member when there isn't Group")
        if isinstance(member, Member) and not group:
            group = member.group
        await self.communicator.send_handle(
            "memberInfo",
            "update",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group,
                "memberId": member.id if isinstance(member, Member) else member,
                "info": {'name': member_info.name, 'specialTitle': member_info.specialTitle}
            }
        )

    async def modify_admin(self, is_admin: bool, member: Union[Member, int], group: Optional[Union[Group, int]] = None):
        if not group and isinstance(member, int):
            raise ValueError("You must give a Member when there isn't Group")
        if isinstance(member, Member) and not group:
            group = member.group
        await self.communicator.send_handle(
            "memberAdmin",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "target": group.id if isinstance(group, Group) else group,
                "memberId": member.id if isinstance(member, Member) else member,
                "assign": is_admin
            }
        )

    async def upload_image(self, data: bytes, may_method: Optional[str] = None, is_flash: bool = False):
        from .message.element import Image
        from .message.element import FlashImage
        method = may_method or upload_method.get().value
        result = await self.communicator.send_handle(
            "uploadIamge",
            "MULIPART",
            {
                "sessionKey": self.bot_session.sessionKey,
                "type": method,
                "img": data
            }
        )
        if is_flash:
            return FlashImage.parse_obj(result)
        return Image.parse_obj(result)

    async def upload_voice(self, data: bytes, may_method: Optional[str] = None):
        from .message.element import Voice
        method = may_method or upload_method.get().value
        if method == "group":
            result = await self.communicator.send_handle(
                "uploadVoice",
                "MULIPART",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "type": method,
                    "img": data
                }
            )
            return Voice.parse_obj(result)
        else:
            raise TypeError("Voice is only provides sending in group!")

    async def upload_file(self, data: bytes, target: Union[Group, int], may_method: Optional[str] = None,
                          path: str = "", ):
        method = may_method or upload_method.get().value
        if method == "group":
            result = await self.communicator.send_handle(
                "file/upload",
                "MULTIPART",
                {
                    "sessionKey": self.bot_session.sessionKey,
                    "type": method.value,
                    "target": target if isinstance(target, int) else target.id,
                    "path": path,
                    "file": data,
                },
            )
            return FileInfo.parse_obj(result)
        else:
            raise TypeError("File is only provides sending in group!")

    async def file_get_list(self, target: Union[Group, int], dir_id: str = "", with_download_info: bool = False,
                            offset: Optional[int] = 0, size: Optional[int] = 1):
        """
        列出指定文件夹下的所有文件.

        Args:
           target (Union[Group, int]): 目标文件的群号
           dir_id (str): 文件ID, 空串为根目录
           with_download_info (bool): 是否携带下载信息, 无必要不要携带
           offset (int): 分页偏移
           size (int): 分页大小
        Returns:
           FileInfo: 返回的文件信息列表.
        """

        result = await self.communicator.send_handle(
            "file/list",
            "GET",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "target": target.id if isinstance(target, Group) else target,
                "withDownloadInfo": str(with_download_info),
                "offset": offset,
                "size": size
            },
        )

        return [FileInfo.parse_obj(i) for i in result]

    async def file_get_info(self, target: Union[Group, int], dir_id: str = "", with_download_info: bool = False):
        """
        获取指定文件的信息.

        Args:
           target (Union[Group, int]): 目标文件的群号
           dir_id (str): 文件ID, 空串为根目录
           with_download_info (bool): 是否携带下载信息, 无必要不要携带
        Returns:
           FileInfo: 返回的文件信息.
        """

        result = await self.communicator.send_handle(
            "file/list",
            "GET",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "target": target.id if isinstance(target, Group) else target,
                "withDownloadInfo": str(with_download_info)
            },
        )
        return FileInfo.parse_obj(result)

    async def file_make_directory(self, target: Union[Group, int], directory_name: str, dir_id: str = ""):
        """
        在指定位置创建新文件夹.

        Args:
            target (Union[Group, int]): 目标文件的群号
            directory_name (str): 新建文件夹名.
            dir_id (str): 父目录id,空串为根目录
        Returns:
            FileInfo: 新创建文件夹的信息.
        """
        result = await self.communicator.send_handle(
            "file/list",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "directoryName": directory_name,
                "target": target.id if isinstance(target, Group) else target,
            },
        )
        return FileInfo.parse_obj(result)

    async def file_delete(self, target: Union[Group, int], dir_id: str = ""):
        """
        删除指定文件.

        Args:
           target (Union[Group, int]): 目标文件的群号
           dir_id (str): 删除文件的id
        """
        await self.communicator.send_handle(
            "file/delete",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "target": target.id if isinstance(target, Group) else target,
            },
        )

    async def file_move(self, target: Union[Group, int], move_to: str, dir_id: str = ""):
        """
        移动指定文件.

        Args:
           target (Union[Group, int]): 目标文件的群号
           move_to (str): 移动目标文件夹的id
           dir_id (str): 移动文件的id
        """
        await self.communicator.send_handle(
            "file/move",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "target": target.id if isinstance(target, Group) else target,
                "moveTo": move_to
            },
        )

    async def file_rename(self, target: Union[Group, int], rename_to: str, dir_id: str = ""):
        """
        重命名指定文件.

        Args:
           target (Union[Group, int]): 目标文件的群号
           rename_to (str): 新文件名
           dir_id (str): 重命名文件id
        """
        await self.communicator.send_handle(
            "file/rename",
            "POST",
            {
                "sessionKey": self.bot_session.sessionKey,
                "id": dir_id,
                "target": target.id if isinstance(target, Group) else target,
                "renameTO": rename_to
            },
        )



import setuptools

with open("README.rst", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="cesloi",
    version="0.6.10",
    author="ArcletProject",
    author_email="rf_tar_railt@qq.com",
    description="An simple Python SDK base on mirai-api-http v2",
    license='AGPL 3.0',
    long_description=long_description,
    long_description_content_type="text/rst",
    url="https://github.com/RF-Tar-Railt/Cesloi",
    install_requires=['aiohttp', 'yarl', 'pydantic', 'loguru', 'arclet-letoderea', 'arclet-alconna'],
    packages=['arclet.cesloi', 'arclet.cesloi.event', 'arclet.cesloi.message', 'arclet.cesloi.model', 'arclet.cesloi.timing'],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    keywords='mirai, bot, asyncio, http, websocket',
    python_requires='>=3.8'
)



from typing import Any, TYPE_CHECKING
from arclet.letoderea.entities.event import TemplateEvent
if TYPE_CHECKING:
    pass


class ApplicationRunning(TemplateEvent):
    bot: Any

    def __init__(self, bot):
        self.bot = bot

    def get_params(self):
        return self.param_export(
            Cesloi=self.bot
        )


class ApplicationStop(TemplateEvent):
    bot: Any

    def __init__(self, bot):
        self.bot = bot

    def get_params(self):
        return self.param_export(
            Cesloi=self.bot
        )


from typing import Union

from arclet.letoderea.breakpoint import StepOut
from .message.messageChain import MessageChain
from .event.messages import GroupMessage, TempMessage, FriendMessage


def group_message_handler(matched_message: Union[str, MessageChain]):

    class GroupWaiter(StepOut):
        def __init__(self):
            super().__init__(GroupMessage)

        @staticmethod
        def handler(message: MessageChain):
            if isinstance(matched_message, str):
                if message.startswith(matched_message):
                    return message
            elif message.to_text() == matched_message.to_text():
                return message

    return GroupWaiter()


def friend_message_handler(matched_message: Union[str, MessageChain]):

    class FriendWaiter(StepOut):
        def __init__(self):
            super().__init__(FriendMessage)

        @staticmethod
        def handler(message: MessageChain):
            if isinstance(matched_message, str):
                if message.startswith(matched_message):
                    return message
            elif message.to_text() == matched_message.to_text():
                return message

    return FriendWaiter()


def temp_message_handler(matched_message: Union[str, MessageChain]):

    class TempWaiter(StepOut):
        def __init__(self):
            super().__init__(TempMessage)

        @staticmethod
        def handler(message: MessageChain):
            if isinstance(matched_message, str):
                if message.startswith(matched_message):
                    return message
            elif message.to_text() == matched_message.to_text():
                return message

    return TempWaiter()



"""
该文件为Cesloi的小型集成库
"""
from .bot_client import Cesloi
from .model.relation import Group, Member, Friend
from ..letoderea.entities.subscriber import Subscriber
from .communicate_with_mah import BotSession
from .message.messageChain import MessageChain
from .event.messages import GroupMessage, FriendMessage, TempMessage


class CesloiMini:
    """
    Cesloi的小型集成库,只选择了基础功能
    """
    def __init__(self, host, qq, verify_key):
        self.client = Cesloi
        self.app = Cesloi(
            bot_session=BotSession(host=host, account=qq, verify_key=verify_key),
            debug=True)
        self.Friend = Friend
        self.Group = Group
        self.Member = Member
        self.Message = MessageChain
        self.GroupMessage = GroupMessage
        self.FriendMessage = FriendMessage
        self.TempMessage = TempMessage

    def register(self, *args, **kwargs):
        """
        注册事件方法，用于指定订阅器订阅的发布器绑定的事件。
        """
        def _wrapper(func):
            return self.app.register(*args, **kwargs)(Subscriber.set()(func))
        return _wrapper

    def run(self):
        return self.app.start()



from .base import MiraiEvent
from arclet.cesloi.model.relation import Client, Friend, Member, Stranger, Sender
from arclet.cesloi.message.messageChain import MessageChain
from .inserter import ApplicationInserter, EventInserter


class Message(MiraiEvent):
    type: str
    messageChain: MessageChain
    sender: Sender

    def __eq__(self, other: "Message"):
        return self.messageChain.to_text() == other

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Sender=self.sender
        )


class FriendMessage(Message):
    type: str = "FriendMessage"
    sender: Friend

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Friend=self.sender
        )


class GroupMessage(Message):
    type: str = "GroupMessage"
    sender: Member

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Member=self.sender,
            Group=self.sender.group
        )


class TempMessage(Message):
    type: str = "TempMessage"
    sender: Member

    def get_params(self,):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Member=self.sender,
            Group=self.sender.group
        )


class StrangerMessage(Message):
    type: str = "StrangerMessage"
    sender: Stranger

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Stranger=self.sender
        )


class OtherClientMessage(Message):
    type: str = "OtherClientMessage"
    sender: Client

    def get_params(self):
        return self.param_export(
            ApplicationInserter,
            EventInserter,
            MessageChain=self.messageChain,
            Client=self.sender
        )


class InvalidEventTypeDefinition(Exception):
    """不合法的事件类型定义."""


class InvalidVerifyKey(Exception):
    """无效的 verifyKey 或其配置."""


class AccountNotFound(Exception):
    """指定的Bot不存在"""


class InvalidSession(Exception):
    """Session失效或不存在"""


class UnVerifiedSession(Exception):
    """Session未认证(未激活)"""


class UnknownTarget(Exception):
    """发送消息目标不存在(指定对象不存在)"""


class AccountMuted(Exception):
    """Bot被禁言，指Bot当前无法向指定群发送消息."""


class MessageTooLong(Exception):
    """消息过长, 尝试分段发送或报告问题."""


class InvalidArgument(Exception):
    """错误的访问，如参数错误等."""


class UnknownError(Exception):
    """其他错误"""

